
import abc, uuid, numpy as np
import asyncio, traceback, sys

from ..JHTML import JHTML

from .Interfaces import Component, ListGroup, Dropdown, Progress
from .Variables import Var, InterfaceVars

__all__ = [
    "Control",
    "InputField",
    "StringField",
    "Slider",
    "Checkbox",
    "RadioButton",
    "Switch",
    "TextArea",
    "Selector",
    "VariableDisplay",
    "FunctionDisplay",
    # "DropdownMenu",
    "MenuSelect",
    "DropdownSelect",
    "ProgressBar"
]

__reload_hook__ = ['.Interfaces', '.Variables']

#region JHTML Controls
class Control(Component):
    layout_orientation = 'row'
    def __init__(self, var, namespace=None):
        self.var = Var(var, namespace=namespace)
        self._widget_cache = None
        super().__init__()
    def to_widget(self, parent=None):
        needs_link = self._widget_cache is None
        widg = super().to_widget(parent=parent)
        if needs_link:
            val = self.var.value
            self.var.link(self)
            self.var.value = val
            self.set_value()
        return widg
    @abc.abstractmethod
    def set_value(self):
        ...
    @abc.abstractmethod
    def get_value(self):
        ...
    @property
    def value(self):
        return self.get_value()
    @value.setter
    def value(self, v):
        self.var.set_value(v, caller=self)
        self.set_value()
    def observe(self, fn, names=None):
        if self._widget_cache is None:
            raise ValueError("not initialized")
        return self._widget_cache.to_widget().observe(fn, names=names)

    control_types = {}
    @classmethod
    def construct(cls, var, control_type=None, field_type=None, value=None, **attrs):
        if control_type is None:
            control_type = cls.infer_control(field_type=field_type, value=value, **attrs)
        if isinstance(control_type, str):
            control_type = cls.control_types[control_type]
        if value is not None:
            value = str(value)
        return control_type(var, value=value, **attrs)
    @classmethod
    def infer_control(cls, field_type=None, value=None, **ignored):
        if field_type is None and value is not None:
            field_type = type(value)
        if field_type is not None:
            if issubclass(field_type, str):
                return cls.control_types['StringField']
            elif issubclass(field_type, (bool,)):
                return cls.control_types['Checkbox']
            elif issubclass(field_type, (int, np.integer, float, np.floating)):
                return cls.control_types['Slider']
            else:
                raise NotImplementedError("can't infer control type for 'field_type' {}".format(field_type))
        elif 'range' in ignored:
            return cls.control_types['Slider']
        else:
            raise NotImplementedError("can't infer control type without 'field_type'")

class ValueWidget(Control):
    def __init__(self, var, value=None):
        super().__init__(var)
        if value is not None:
            self.var.value = value
        if self.var.value is None:
            self.var.value = ""
    def get_value(self):
        if self._widget_cache is not None:
            val = self._widget_cache.value
            return "" if val is None else val
        else:
            return self.var.value
    def set_value(self):
        if self._widget_cache is not None:
            self._widget_cache.value = self.var.value
    def update(self, e):
        if self._widget_cache is not None:
            self.var.value = self._widget_cache.value
class InputField(ValueWidget):
    base_cls = ['form-control']
    def __init__(self, var, value=None, tag='input', track_value=True, continuous_update=False, base_cls=None, cls=None, **attrs):
        super().__init__(var, value=value)
        if base_cls is not None:
            self.base_cls = base_cls
        attrs['cls'] = self.base_cls + JHTML.manage_class(cls)
        attrs['tag'] = tag
        attrs['track_value'] = track_value
        attrs['continuous_update'] = continuous_update
        self.attrs = attrs
    def to_jhtml(self):
        # value = self.var.value
        # if value is not None and not isinstance(value, str):
        #     value = str(value)
        # self._attrs['value'] = value
        field = JHTML.Input(**self.attrs)
        return field
class StringField(InputField):
    # base_cls = ['form-text']
    def __init__(self, var, type='text', **attrs):
        super().__init__(var, type=type, **attrs)
Control.control_types['StringField'] = StringField
class Slider(InputField):
    base_cls = ['form-range']
    def __init__(self, var, type='range', value=None, range=None, **attrs):
        if range is not None:
            if value is None:
                value = range[0]
            min = attrs.get('min', None)
            if min is None:
                min = range[0]
            max = attrs.get('max', None)
            if max is None:
                max = range[1]
            step = attrs.get('step', None)
            if step is None:
                step = (max-min) / 25 if len(range) == 2 else range[2]
            attrs.update(
                min=min,
                max=max,
                step=step
            )
        if value is not None and not isinstance(value, str):
            value = str(value)
        super().__init__(var, type=type, value=value, **attrs)
    def get_value(self):
        if self._widget_cache is not None:
            val = self._widget_cache.value
            try:
                val = int(val)
            except (ValueError, TypeError):
                try:
                    val = float(val)
                except (ValueError, TypeError):
                    ...
            return val
    def set_value(self):
        if self._widget_cache is not None:
            v = self.var.value
            if not isinstance(v, str):
                v = str(v)
            self._widget_cache.value = v
Control.control_types['Slider'] = Slider
class Checkbox(InputField):
    base_cls = ['form-check-input']
    def __init__(self, var, type='checkbox', **attrs):
        super().__init__(var, type=type, **attrs)
    def get_value(self):
        if self._widget_cache is not None:
            val = self._widget_cache.value
            return isinstance(val, str) and val == "true"
    def set_value(self):
        if self._widget_cache is not None:
            if self.var.value:
                self._widget_cache.value = 'true'
            else:
                self._widget_cache.value = 'false'
Control.control_types['Checkbox'] = Checkbox
class RadioButton(Checkbox):
    base_cls = ['form-check-input']
    def __init__(self, var, type='radio', **attrs):
        super().__init__(var, type=type, **attrs)
Control.control_types['RadioButton'] = RadioButton
class Switch(Checkbox):
    base_cls = ['form-check-input']
    def __init__(self, var, type='checkbox', role='switch', **attrs):
        super().__init__(var, type=type, role=role, **attrs)
    def get_value(self):
        if self._widget_cache is not None:
            val = self._widget_cache.get_child(0).value
            return isinstance(val, str) and val == "true"
    def set_value(self):
        if self._widget_cache is not None:
            if self.var.value:
                self._widget_cache.get_child(0).value = 'true'
            else:
                self._widget_cache.get_child(0).value = 'false'
    def to_jhtml(self):
        return JHTML.Div(super().to_jhtml(), cls=['form-switch'])
Control.control_types['Switch'] = Switch
class TextArea(InputField):
    base_cls = ['form-control']
    def __init__(self, var, tag='textarea', **attrs):
        super().__init__(var, tag=tag, **attrs)
    def to_jhtml(self):
        field = JHTML.Textarea(**self.attrs)
        return field
Control.control_types['TextArea'] = TextArea
class ChangeTracker(ValueWidget):
    base = None
    base_cls = []
    def __init__(self, var, base=None, value=None, track_value=True, continuous_update=False, base_cls=None, cls=None, **attrs):
        super().__init__(var, value=value)
        base = self.base if base is None else base
        self.base = getattr(JHTML, base) if isinstance(base, str) else base
        if base_cls is not None:
            self.base_cls = base_cls
        attrs['cls'] = self.base_cls + JHTML.manage_class(cls)
        attrs['track_value'] = track_value
        attrs['continuous_update'] = continuous_update
        self.attrs = attrs
class Selector(ChangeTracker):
    base=JHTML.Select
    base_cls = ['form-select']
    def __init__(self, var, options=None, value=None, multiple=False, **attrs):
        self._options = self.canonicalize_options(options)
        if not multiple and value is None and len(self._options) > 0:
            value = self._options[0][1]
        if multiple:
            attrs['multiple']=True
        super().__init__(var, value=value, **attrs)
    @property
    def multiple(self):
        if self._widget_cache is not None:
            try:
                mult = self._widget_cache['multiple']
            except KeyError:
                mult = False
            return mult is True or mult == 'true'
        else:
            return 'multiple' in self.attrs and self.attrs['multiple']
    def get_value(self):
        if self._widget_cache is not None:
            val = super().get_value()
            if self.multiple:
                if val is None or val == "":
                    val = []
                else:
                    val = val.split("&&")
            elif val == "":
                val = None
            return val
    def set_value(self):
        if self._widget_cache is not None:
            if self.multiple:
                v = self.var.value
                if not isinstance(v, str):
                    v = "&&".join(v)
                self._widget_cache.value = v
            else:
                super().set_value()
    @classmethod
    def canonicalize_options(cls, options):
        ops = []
        for k in options:
            if isinstance(k, str):
                v = k
            else:
                try:
                    k, v = k
                except ValueError:
                    v = k
            ops.append((k,v))
        return tuple(ops)
    def _build_options_list(self):
        if self.multiple:
            vals = self.var.value
            if vals is None:
                vals = []
            elif isinstance(vals, str):
                vals = vals.split("&&")
            opts = [JHTML.Option(k, value=v, selected="true") if v in vals else JHTML.Option(k, value=v) for k,v in self._options]
        else:
            val = self.var.value
            opts = [JHTML.Option(k, value=v, selected=(val is not None and v == val)) for k,v in self._options]
        return opts
    def to_jhtml(self):
        field = self.base(*self._build_options_list(), **self.attrs)
        return field
Control.control_types['Selector'] = Selector
class VariableDisplay(Control):
    def __init__(self, var, value=None, pane=None, autoclear=True, namespace=None, **attrs):
        super().__init__(var, namespace=namespace)
        if pane is None:
            pane = JHTML.OutputArea(autoclear=autoclear, **attrs)
        self.out = pane
        if value is not None:
            self.var.value = value
    def get_value(self):
        return self.var.value
    def set_value(self):
        val = self.var.value
        if val is None or isinstance(val, str) and val == "":
            self.out.clear()
        else:
            self.out.set_output(val)
    def update(self, e):
        self.set_value()
    def to_jhtml(self):
        with self.out:
            self.set_value()
        return self.out
class FunctionDisplay(Component):
    def __init__(self, fn, vars, pane=None, autoclear=True, debounce=None, **attrs):
        super().__init__()
        if pane is None:
            pane = JHTML.OutputArea(autoclear=autoclear, **attrs)
        self.out = pane
        self.fn = fn
        self.debounce = debounce
        self._delayed_executor = None
        # self._executions = []
        self.vars = InterfaceVars(*vars) if not isinstance(vars, InterfaceVars) else vars
    def link_vars(self, *var):
        self.update = self.update  # weakref patch
        for v in self.vars:
            v.callbacks.add(self.update)
            v.link(self)
        self.vars.callbacks.add(self.link_vars)
    def to_widget(self, parent=None):
        needs_link = self._widget_cache is None
        widg = super().to_widget(parent=parent)
        if needs_link:
            self.link_vars()
        return widg
    def observe(self, fn, names=None):
        if self._widget_cache is None:
            raise ValueError("not initialized")
        return self._widget_cache.to_widget().observe(fn, names=names)

    # directly from the jupyter docs
    async def _delayed_update(self, event):
        await asyncio.sleep(self.debounce)
        return self._update(event)
    def update(self, event):
        try:
            if self.debounce is not None:
                if self._delayed_executor is not None:
                    self._delayed_executor.cancel()
                self._delayed_executor = asyncio.ensure_future(self._delayed_update(event))
            else:
                self._update(event)
        except:
            with self.out:
                _, _, tb = sys.exc_info()
                traceback.print_tb(tb)

    def _update(self, e):
        res = self.fn(event=e, pane=self, **self.vars.dict)
        self._last_res = res
        if res is not None:
            self.out.set_output(res)
            self._delayed_executor = None

    def to_jhtml(self):
        with self.out:
            self._update(None)
        return self.out

class ProgressBar(Control):
    def __init__(self, var, bar=None, **attrs):
        super().__init__(var)
        if bar is None:
            bar = Progress(self.value, **attrs)
        self.bar = bar
    def get_value(self):
        val = self.var.value
        if val is None or val == "":
            self.var.value = 0
        elif isinstance(val, str):
            self.var.val = int(val)
        return self.var.value
    def set_value(self):
        self.bar.bar['width'] = str(self.value) + "%"
    def update(self, e):
        self.set_value()
    def to_jhtml(self):
        self.set_value()
        return self.bar.to_widget()

class MenuSelect(ValueWidget):
    menu_type = ListGroup
    def __init__(self, var, options, menu_type=None, **attrs):
        super().__init__(var)
        self._value_map, self.ops = self.canonicalize_options(options)
        if menu_type is not None:
            self.menu_type = menu_type
        self._obj = self.menu_type(self.ops, **attrs) #type: MenuComponent
        self._active_item = None
    def get_value(self):
        if self._active_item is not None:
            return self._value_map[self._active_item]
    def set_value(self):
        self.set_active(self.var.value)
    def update(self, e):
        self.set_active(self.value)
    def set_active(self, v):
        for k, test in self._value_map.items():
            if test == v:
                break
        else:
            k = None
        if k is not None:
            self.set_active_key(k)
    def set_active_key(self, k):
        if len(self._obj._item_map) > 0:
            if (
                    self._active_item is not None
                    and k != self._active_item
            ):
                cur = self._obj._item_map[self._active_item]
                if cur is not None:
                    cur.remove_class('active')
            self._active_item = k
            new = self._obj._item_map[k]
            if new is not None:
                new.add_class('active')
    def onclick(self, e, i, v):
        self.var.set_value(v, caller=self)
        # with self.logger_pane:
        self.set_active(v)

    def canonicalize_options(self, options):
        ops = []
        val_dict = {}
        for k in options:
            if not isinstance(k, dict):
                if isinstance(k, str):
                    v = k
                else:
                    try:
                        k, v = k
                    except (TypeError, ValueError):
                        v = k
                k = {
                    'body':k,
                    'value':v
                }
            if 'id' not in k:
                uid = str(uuid.uuid4()).replace("-", "")
                k['id'] = uid
            val_dict[k['id']] = k['value']
            k['event_handlers'] = {'click':lambda *e,i=k['id'],v=k['value']:self.onclick(e, i, v)}
            ops.append(k)
        return val_dict, tuple(ops)
    def to_jhtml(self):
        widg = self._obj.to_jhtml()
        init_key = next(iter(self._value_map.keys()))
        self.set_active(init_key)
        return widg

class DropdownSelect(ValueWidget):
    menu_type = Dropdown
    def __init__(self, var, options, name=None, menu_type=None, **attrs):
        super().__init__(var)
        if menu_type is not None:
            self.menu_type = menu_type
        self.selector = MenuSelect(var, options, menu_type=self.menu_type.List)
        if name is None:
            name = self.var.name
        self.name = name
        self.attrs = attrs

    def get_value(self):
        return self.selector.get_value()
    def set_value(self):
        self.selector.set_value()
    def update(self, e):
        self.selector.update(e)
    def to_jhtml(self):
        return self.menu_type(
            self.name,
            self.selector.to_widget(),
            **self.attrs
        ).to_jhtml()



import abc, weakref
import enum
from collections import OrderedDict
from .. import Devutils as dev

__all__ = [
    "Cache",
    "MaxSizeCache",
    "ObjectRegistry"
]

class Cache(metaclass=abc.ABCMeta):
    """
    Simple cache base class
    """
    @abc.abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError("Cache is an abstract base class")
    def get(self, item, default=None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default
    @abc.abstractmethod
    def __contains__(self, item):
        raise NotImplementedError("Cache is an abstract base class")
    @abc.abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError("Cache is an abstract base class")

class MaxSizeBackend(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def keys(self):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")
    @abc.abstractmethod
    def __len__(self):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")
    @abc.abstractmethod
    def __getitem__(self, item):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")
    @abc.abstractmethod
    def __contains__(self, item):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")
    @abc.abstractmethod
    def __setitem__(self, key, value):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")
    @abc.abstractmethod
    def pop(self):
        cls = type(self)
        raise NotImplementedError(f"{cls.__name__} is an abstract base class")

class LRUDict(MaxSizeBackend):
    def __init__(self):
        self.od = OrderedDict()
    def __len__(self):
        return len(self.od)
    def keys(self):
        return self.od.keys()
    def __contains__(self, item):
        return item in self.od
    def __getitem__(self, item):
        val = self.od[item]
        self.od.move_to_end(item)
        return val
    def __setitem__(self, key, value):
        self.od[key] = value
        self.od.move_to_end(key)
    def pop(self):
        return self.od.popitem(last=False)

class FIFODict(MaxSizeBackend):
    def __init__(self):
        self.od = {}
    def __len__(self):
        return len(self.od)
    def keys(self):
        return self.od.keys()
    def __contains__(self, item):
        return item in self.od
    def __getitem__(self, item):
        val = self.od[item]
        return val
    def __setitem__(self, key, value):
        self.od[key] = value
    def pop(self):
        return self.od.pop(next(iter(self.od.keys())))

class MaxSizeCache(Cache):
    """
    Simple lru-cache to support ravel/unravel ops
    """
    def __init__(self, max_items=128, cache_type=None):
        self.backend = self.resolve_cache_type(cache_type)
        self.max_items = max_items
    class Backends(enum.Enum):
        LRU = "lru"
        FIFO = "fifo"
    cache_types = dev.OptionsMethodDispatch(
        {
            Backends.LRU:LRUDict,
            Backends.FIFO:FIFODict,
        },
        methods_enum=Backends,
        default_method=Backends.LRU
    )
    @classmethod
    def resolve_cache_type(cls, type_name):
        if callable(type_name):
            method = type_name
            opts = {}
        else:
            method, opts = cls.cache_types.resolve(type_name)
        return method(**opts)

    def keys(self):
        return self.backend.keys()
    def __contains__(self, item):
        return item in self.backend
    def __getitem__(self, item):
        return self.backend.__getitem__(item)
    def __setitem__(self, key, value):
        self.backend.__setitem__(key, value)
        if len(self.backend) > self.max_items:
            self.backend.pop()

class ObjectRegistryDefaults:
    Raise="raise"
    NotFound="NotFound"

class RegistryDefaultContext:
    def __init__(self, registry:'ObjectRegistry', value):
        self.reg = registry
        self.old = None
        self.val = value
    def __enter__(self):
        self.old = self.reg.default
        self.reg.default = self.val
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reg.default = self.old

class ObjectRegistry:
    """
    Provides a simple interface to global object registries
    so that pieces of code don't need to pass things like loggers
    or parallelizers through every step of the code
    """

    def __init__(self, default=ObjectRegistryDefaults.Raise):
        self.cache = weakref.WeakValueDictionary() # allow unused vals to get gc'd
        self.default = default

    def temp_default(self, val):
        return RegistryDefaultContext(self, val)

    def __contains__(self, item):
        return self.cache.__contains__(item)

    def lookup(self, key):
        if self.default is ObjectRegistryDefaults.Raise:
            return self.cache[key]
        else:
            try:
                return self.cache[key]
            except KeyError:
                return self.default
    def __getitem__(self, item):
        return self.lookup(item)

    def register(self, key, val):
        self.cache[key] = val
    def __setitem__(self, key, value):
        self.register(key, value)

    def keys(self):
        return self.cache.keys()
    def items(self):
        return self.cache.items()
    def values(self):
        return self.cache.values()


"""
A package for managing interfaces with electronic structure packages.
Provides a generic set of properties and a job template interface.
"""


__all__ = []
from .Parsers import *; from .Parsers import __all__ as exposed
__all__ += exposed
from .Orca import *; from .Orca import __all__ as exposed
__all__ += exposed
from .GaussianImporter import *; from .GaussianImporter import __all__ as exposed
__all__ += exposed
from .FChkDerivatives import *; from .FChkDerivatives import __all__ as exposed
__all__ += exposed
from .CIFParser import *; from .CIFParser import __all__ as exposed
__all__ += exposed
from .Crest import *; from .Crest import __all__ as exposed
__all__ += exposed



import io
import base64
from .Interface import *

__all__ = [
    "Open3DInterface",
]

class Open3DInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenCV toolkit
    """
    name = 'Open3D'
    module = 'open3d'



import abc, numpy as np, io, weakref
from .. import Devutils as dev
from .. import Numputils as nput
from .TableFormatters import TableFormatter
from ..Misc.Symbolics import Abstract

__all__ = [
    "TeX"
]

class TeXWriter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def format_tex(self, context=None):
        ...

    real_digits = 3
    @classmethod
    def dispatch_format(cls, b, context):
        if isinstance(b, TeXWriter) or hasattr(b, 'format_tex'):
            return b.format_tex(context)
        elif isinstance(b, (float, np.floating)):
            return ('{:.'+str(cls.real_digits)+'f}').format(b)
        elif isinstance(b, np.ndarray):
            return TeXArray(b).format_tex(context)
        elif isinstance(b, (list, tuple)):
            if isinstance(b[0], (list, tuple)):
                return TeXArray(b).format_tex(context)
            else:
                return TeXRow(b).format_tex(context)
        else:
            return str(b)

    def as_expr(self):
        return TeXExpr(self)

class TeXContextManager:
    default_contexts = weakref.WeakValueDictionary()
    @classmethod
    def resolve(cls, name='default'):
        if name not in cls.default_contexts:
            ctx = cls()
            cls.default_contexts[name] = ctx
        return cls.default_contexts[name]

    def __init__(self):
        self.context_stack = []
    def subcontext(self, cls):
        return cls(self)
    def set_context(self, ctx):
        self.context_stack.append(ctx)
        return ctx
    def leave_context(self):
        self.context_stack.pop()
    @property
    def context(self):
        if len(self.context_stack) == 0:
            return None
        else:
            return self.context_stack[-1]
    @property
    def math_mode(self):
        return isinstance(self.context, MathContext)

class TeXContext:
    def __init__(self, manager:TeXContextManager):
        self.manager = manager
    def __enter__(self):
        self.manager.set_context(self)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.manager.leave_context()
class MathContext(TeXContext):
    ...

class TeXBlock(TeXWriter):
    tag = None
    modifier = None
    modifier_type = '[]'
    separator = '\n'
    context = None
    label_header = None
    def __init__(self,
                 body=None, *,
                 tag=None,
                 modifier=None,
                 modifier_type=None,
                 separator=None,
                 context=None,
                 label=None
                 ):
        self.body = body
        if tag is None:
            tag = self.tag
        self.tag = tag
        if modifier is None:
            modifier = self.modifier
        self.modifier = modifier
        if modifier_type is None:
            modifier_type = self.modifier_type
        self.modifier_type = modifier_type
        if separator is None:
            separator = self.separator
        self.sep = separator
        if context is None:
            context = self.context
        self.ctx = context
        self.label = label
    def prep_body(self, context=None):
        if self.body is None:
            params = []
        elif isinstance(self.body, (list, tuple)):
            params = [self.dispatch_format(b, context) for b in self.body]
        else:
            params = [self.dispatch_format(self.body, context)]
        if self.label is not None:
            label = self.label
            if self.label_header is not None and ':' not in label:
                label = self.label_header + ':' + label
            params = params + ['\\label{' + label + "}"]
        return params
    @classmethod
    def construct_modified_tag(self, tag, mod, mod_type='[]'):
        header = "\\begin{"+str(tag)+"}"
        if mod is not None:
            if isinstance(mod_type, str):
                l, r = mod_type
                header = header + l + mod + r
            else:
                bits = []
                for m,t in zip(mod, mod_type):
                    l, r = t
                    bits.append(l + m + r)
                header = header + "".join(bits)
        return header, "\\end{"+str(tag)+"}"
    def construct_header_footer(self):
        return self.construct_modified_tag(self.tag, self.modifier, self.modifier_type)
    def format_body(self, body_params):
        header, footer = self.construct_header_footer()
        if self.tag is not None:
            body_params = [header] + body_params + [footer]
        return self.sep.join(body_params)
    def format_tex(self, context=None):
        if self.ctx is not None:
            if context is None:
                context = TeXContextManager.resolve()
            elif isinstance(context, str):
                context = TeXContextManager.resolve(context)

        if self.ctx is not None:
            with context.subcontext(self.ctx):
                body_args = self.prep_body(context)
        else:
            body_args = self.prep_body(context)
        return self.format_body(body_args)
    def __call__(self, body):
        return type(self)(
            body,
            tag=self.tag,
            modifier=self.modifier,
            modifier_type=self.modifier_type,
            separator=self.separator,
            context=self.ctx,
            label=self.label
        )

class TeXRow(TeXBlock):
    tag = None
    separator = ' '

class TeXArray(TeXBlock):
    tag = 'tabular'
    modifier_type = '{}'
    separator = '\n'
    array_separator = " & "
    array_newline = " \\\\\n"
    header_separator = "\n\\hline \\\\[-4ex]\n"
    header_lines = "\n\\hline \\\\[-4ex]"
    footer_lines = "\\hline \\\\[-4ex]\n"
    number_format = "{:8.3f}"
    def __init__(self, headers_or_body, body=None, *,
                 alignment='auto',
                 number_format="{:8.3f}",
                 content_join=None,
                 column_join=None,
                 row_join=None,
                 separator=None,
                 header_spans=None,
                 header_alignments=None,
                 resizeable=False,
                 **opts):
        self.resizeable = resizeable
        if resizeable:
            self.tag = 'tabularx'
            self.modifier_type = ["{}", "{}"]
            self.array_separator = " && "

        if body is None:
            body = headers_or_body
            headers_or_body = None
        self.headers = headers_or_body
        self.alignment = alignment
        if content_join is None:
            content_join = self.header_separator
        opts['content_join'] = content_join
        if row_join is None:
            row_join = self.array_newline
        opts['row_join'] = row_join
        if column_join is None:
            column_join = self.array_separator
        opts['column_join'] = column_join
        if separator is None:
            separator = ""
        opts['separator'] = separator
        self.format_opts, opts = dev.OptionsSet(opts).split(TableFormatter)
        self.format = number_format
        self.header_spans = header_spans
        self.header_alignments = header_alignments
        super().__init__(
            body,
            **opts
        )

    def construct_alignment_spec(self, body):
        if isinstance(body, np.ndarray):
            if np.issubdtype(body.dtype, np.integer):
                spec = "c" * body.shape[1]
            elif np.issubdtype(body.dtype, np.floating):
                spec = 'r' * body.shape[1]
            else:
                spec = 'c' * body.shape[1]
        else:
            specs = []
            for array_row in body:
                for i, e in enumerate(array_row):
                    if i >= len(specs):
                        specs = specs + ['r']
                    if (
                            specs[i] != 'c' and
                            not isinstance(e, (float, np.floating)) and
                            not (isinstance(e, str) and len(e.strip()) == 0)
                    ):
                        specs[i] = 'c'
            spec = "".join(specs)
        return spec
    def construct_header_footer(self):
        body = self.body
        if isinstance(body, np.ndarray) and not np.issubdtype(body.dtype, (np.integer, np.floating)):
            body = body.tolist()
        if self.alignment is not None and self.modifier is None:
            mod = self.construct_alignment_spec(body)
        else:
            mod = self.modifier
        if self.resizeable:
            mod = ['\\textwidth', "X".join(mod)]
        header, footer = self.construct_modified_tag(self.tag, mod, self.modifier_type)
        return header + self.header_lines, self.footer_lines + footer

    def format_numpy_array(self, array):
        int_digits = int(np.floor(np.log10(np.max(np.abs(array))))) + 1
        with io.StringIO() as stream:
            if np.issubdtype(array.dtype, np.floating):
                real_digits = self.real_digits
            else:
                real_digits = 0
            total_digits = int_digits + real_digits + 2
            fmt = '%{}.{}f'.format(total_digits, real_digits)
            np.savetxt(stream, array, fmt=fmt, delimiter=self.array_separator, newline=self.array_newline)
            stream.seek(0)
            return stream.read()
    def format_mixed_array(self, array, context=None):
        row_padding = []
        string_array = []
        for array_row in array: # convert and track padding
            conv_row = []
            for i,c in enumerate(array_row):
                s = self.dispatch_format(c, context)
                conv_row.append(s)
                if i >= len(row_padding):
                    row_padding = row_padding + [0]
                if len(s) > row_padding[i]:
                    row_padding[i] = len(s)
            string_array.append(conv_row)
        return self.array_newline.join(
            self.array_separator.join(" " * (row_padding[i] - len(s)) + s for i,s in enumerate(string_row))
            for string_row in string_array
        )
    def prep_body(self, context=None, headers=None, body=None):
        if body is None:
            body = self.body
        if headers is None:
            headers = self.headers
        if headers is None:
            if len(body) == 2 and not all(dev.is_atomic(b) for b in body[0]):
                headers, body = body
        opts = self.format_opts
        if 'column_formats' in opts:
            opts = opts.copy()
            column_formats = opts.pop('column_formats')
        else:
            column_formats = [
                ""
                    if not nput.is_numeric(o) else
                "{:>.0f}"
                    if nput.is_int(o) else
                self.format
                for o in body[0]
            ]

        if headers is not None:
            if dev.is_list_like(headers[0]):
                headers = [
                    [
                        h.format_tex(context=context)
                        if isinstance(h, TeXWriter) else
                        h
                        for h in hl
                    ]
                    for hl in headers
                ]
                if self.header_spans is not None:
                    alignments = self.header_alignments
                    if alignments is None:
                        alignments = [
                            ["c"] * len(hl)
                            for hl in headers
                        ]

                    _blocks = []
                    for lhl, lhs, lhc in zip(headers, self.header_spans, alignments):
                        _ = []
                        for hl, hs, hc in zip(lhl, lhs, lhc):
                            _.append(
                                TeXMulticolumn(2*hs - 1, hc, hl).format_tex(context)
                                    if hs > 1 else
                                hl
                            )
                        _blocks.append(_)
                    headers = _blocks
            else:
                headers = [
                    h.format_tex(context=context)
                        if isinstance(h, TeXWriter) else
                    h
                    for h in headers
                ]
                if self.header_spans is not None:
                    alignments = self.header_alignments
                    if alignments is None:
                        alignments = ["c"] * len(headers)

                    _ = []
                    for hl, hs, hc in zip(headers, self.header_spans, alignments):
                        _.append(
                            TeXMulticolumn(hs, hc, hl).format_tex(context)
                                if hs > 1 else
                            hl
                        )
                    headers = _

        wtf = TableFormatter(
            column_formats,
            headers=headers,
            header_spans=self.header_spans,
            **self.format_opts
        ).format(body) + self.array_newline.strip()
        return [
            wtf
        ]

class TeXTable(TeXBlock):
    tag = 'table'
    modifier = 'ht'
    modifier_type = '[]'
    separator = '\n'
    # array_separator = " & "
    # array_newline = " \\\\\n"
    def __init__(self,
                 headers_or_body,
                 body=None,
                 width=1,
                 caption=None,
                 # label=None,
                 resizeable=False,
                 number_format=None,
                 header_spans=None,
                 **etc
                 ):
        if body is None:
            body = headers_or_body
            headers_or_body = None
        body = [
            TeXArray(headers_or_body, body,
                     number_format=number_format,
                     resizeable=resizeable,
                     header_spans=header_spans
                     )
                if not isinstance(body, TeXWriter) else
            body
        ]
        self.width = width
        self.caption = caption
        # self.label = label
        super().__init__(body, **etc)

    def prep_body(self, context=None, body=None):
        if body is None:
            body = self.body

        base = [
            TeXBlock(
                TeXBlock(
                    body,
                    tag='minipage',
                    modifier_type=["[]", "{}"],
                    modifier=['c', f'{self.width} \\textwidth']
                ),
                tag="center"
            )
        ]
        if self.caption is not None:
            base.append(
                TeXFunction(self.caption, function_name="caption")
                    if not isinstance(self.caption, TeXWriter) else
                self.caption
            )
        if self.label is not None:
            base.append(
                TeXFunction(self.label, function_name="label")
                    if not isinstance(self.label, TeXWriter) else
                self.label
            )
        return [
            b.format_tex(context=context)
            for b in base
        ]

class TeXFunction(TeXWriter):
    function_name = None
    def __init__(self, *args, function_name=None):
        if function_name is None:
            function_name = self.function_name
        self.function_name = function_name
        self.args = args
    def format_tex(self, context=None):
        tag = "\\" + self.function_name
        body = ["{" + self.dispatch_format(b, context) + "}" for b in self.args]
        return tag + "".join(body)

class TeXMulticolumn(TeXFunction):
    function_name = 'multicolumn'
    def __init__(self, width, fmt, body):
        super().__init__(width, fmt, body)

class TeXBold(TeXFunction):
    function_name = 'textbf'

class TeXBracketed(TeXWriter):
    brackets = (None, None)
    def __init__(self, body, brackets=None):
        self.body = body
        if brackets is None:
            brackets = self.brackets
        self.brackets = brackets
    def format_tex(self, context=None):
        l, r = self.brackets
        base = self.dispatch_format(self.body, context)
        return l + base + r

class TeXParenthesized(TeXBracketed):
    brackets = ('\\left(', '\\right)')

class TeXEquation(TeXBlock):
    tag = 'equation'
    context = MathContext
    label_header = 'eq'

########################################################################################################################
#
#       TeX Equations
#
#

#region Equations
class TeXNode(Abstract.Expr):
    def to_ast(self):
        raise NotImplementedError("TeXNodes are for formatting only")
class TeXSuperscript(TeXNode):
    __tag__ = "Superscript"
    __slots__ = ['obj', 'index']
    def __init__(self, obj, index):
        self.obj = obj
        self.index = index
class TeXApply(TeXNode):
    __tag__ = "Apply"
    __slots__ = ['function', 'argument']
    def __init__(self, function, argument):
        self.function = function
        self.argument = argument
class TeXSymbol(Abstract.Name):
    def __call__(self, *args):
        return TeXApply(self, args)

class TeXExpr(TeXWriter):

    @classmethod
    def name(cls, s):
        if isinstance(s, str) and len(s) > 1:
            s = "\\" + s
        return cls(Abstract.Name(s))
    @classmethod
    def symbol(cls, s):
        if isinstance(s, str) and len(s) > 1:
            s = "\\" + s
        return cls(TeXSymbol(s))

    # a, b, c, f, g, i, j, k, l, m, n, x, y, z = Abstract.vars(
    #     'a', 'b', 'c', 'f', 'g', 'i', 'j', 'k', 'l', 'm', 'n',
    #     'x', 'y', 'z'
    # )
    # omega, nu, tau, psi, phi, sigma = Abstract.vars(
    #     '\\omega', '\\nu', '\\tau', '\\psi', '\\phi', '\\sigma'
    # )
    # Omega, Nu, Tau, Psi, Phi, Sigma = Abstract.vars(
    #     '\\Omega', '\\Nu', '\\Tau', '\\Psi', '\\Phi', '\\Sigma'
    # )
    # sum, int, prod, bra, ket, braket = Abstract.vars(
    #     '\\sum', '\\int', '\\prod',
    #     '\\bra', '\\ket', '\\braket',
    #     symbol_type=TeXSymbol
    # )

    def __init__(self, body):
        if not isinstance(body, Abstract.Expr):
            body = Abstract.Name(body)
        self.body = body
    def __add__(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body + other)
    def __radd__(self, other):
        return type(self)(other + self.body)
    def __mul__(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body * other)
    def __rmul__(self, other):
        return type(self)(other * self.body)
    def __pow__(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body ** other)
    def __neg__(self):
        return type(self)(-self.body)
    def __xor__(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body ^ other)
    def __or__(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body | other)
    def __getitem__(self, item):
        return type(self)(self.body[item])

    def Equals(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body.Equals(other))
    Eq = Equals
    def LessThan(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body.LessThan(other))
    Lt = LessThan
    def LessEquals(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body.LessEquals(other))
    LtE = LessEquals
    def GreaterThan(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body.GreaterThan(other))
    Gt = GreaterThan
    def GreaterEquals(self, other):
        if isinstance(other, TeXExpr):
            other = other.body
        return type(self)(self.body.GreaterEquals(other))
    GtE = GreaterEquals

    @staticmethod
    def convert_name(name, converter):
        name = name.name
        if hasattr(name, 'format_tex'):
            name = name.format_tex()
        return name
    @staticmethod
    def convert_const(const, converter):
        return const.value
    @staticmethod
    def convert_call(call, converter):
        return "{}({})".format(
            converter(call.fn),
            ",".join(converter(k) for k in call.args)
        )
    @staticmethod
    def convert_superscript(op, converter):
        return "{}^{{{}}}".format(
            converter(op.obj),
            converter(op.index)
        )
    @staticmethod
    def convert_bitxor(op, converter):
        return "{}^{{{}}}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_power(op, converter):
        return "{}^{{{}}}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_subscript(op, converter):
        idx = op.index
        if isinstance(idx, slice):
            var = idx.start
            min = idx.stop
            max = idx.step
            if max is None:
                return "{}_{{{}={}}}".format(
                    converter(op.obj),
                    converter(var),
                    converter(min)
                )
            else:
                return "{}_{{{}={}}}^{{{}}}".format(
                    converter(op.obj),
                    converter(var),
                    converter(min),
                    converter(max)
                )
        else:
            return "{}_{{{}}}".format(
                converter(op.obj),
                converter(op.index)
            )
    @staticmethod
    def convert_add(op, converter):
        return "{}+{}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_sub(op, converter):
        return "{}-{}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_mul(op, converter):
        return "{} {}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_bitor(op, converter):
        return "{} {}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_div(op, converter):
        return "\\frac{{{}}{{{}}}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_eq(op, converter):
        return "{} = {}".format(
            converter(op.left),
            converter(op.right)
        )
    @staticmethod
    def convert_raw(obj, converter):
        if hasattr(obj, 'format_tex'):
            return obj.format_tex()
        else:
            return TeXBlock(obj, separator=", ").format_tex()

    @property
    def converter_dispatch(self):
        return {
            'Name':self.convert_name,
            'Const':self.convert_const,
            'Superscript':self.convert_superscript,
            'Subscript':self.convert_subscript,
            'Pow':self.convert_power,
            'BitXOr':self.convert_bitxor,
            'BitOr':self.convert_bitor,
            'Add':self.convert_add,
            'Sub':self.convert_sub,
            'Mul':self.convert_mul,
            'Div':self.convert_div,
            'Equals':self.convert_eq,
            None:self.convert_raw
        }

    def format_tex(self, context=None):
        if context is None:
            context = TeXContextManager.resolve()
        elif isinstance(context, str):
            context = TeXContextManager.resolve(context)

        pad_dollars = not context.math_mode
        with context.subcontext(MathContext):
            expr = self.body.transmogrify(
                self.converter_dispatch
            )
        if pad_dollars:
            expr = '${}$'.format(expr)

        return expr

#endregion

########################################################################################################################
#
#       Wrapper
#
#

class TeX:
    """
    Namespace for TeX-related utilities, someday might help with document prep from templates
    """

    Writer = TeXWriter

    Block = TeXBlock
    Row = TeXRow

    Expr = TeXExpr
    Symbol = TeXExpr.name
    Function = TeXExpr.symbol

    Array = TeXArray
    Table = TeXTable
    Equation = TeXEquation

    wrap_parens = TeXParenthesized

    bold = TeXBold

    @classmethod
    def Matrix(cls, mat, **kwargs):
        return cls.wrap_parens(cls.Array(mat, **kwargs))



from .NodeCommServer import ShellCommHandler

__all__ = [
    "GitHandler"
]

class GitHandler(ShellCommHandler):

    DEFAULT_CONNECTION = None
    DEFAULT_PORT_ENV_VAR = 'GIT_SOCKET_PORT'
    # DEFAULT_CONNECTION = os.path.expanduser("~/.gitsocket")
    def get_methods(self) -> 'dict[str,method]':
        return {
            'git':self.do_git
        }
    def do_git(self, args, kwargs):
        return self.subprocess_response("git", args)


import enum
from mmap import mmap
import abc, io
import sys
import textwrap

__all__ = [
    "FileStreamReader",
    "FileStreamCheckPoint",
    "FileStreamerTag",
    "FileStreamReaderException",
    "StringStreamReader",
    "FileLineByLineReader",
    "StringLineByLineReader"
]

########################################################################################################################
#
#                                           FileStreamReader
#
class FileStreamCheckPoint:
    """
    A checkpoint for a file that can be returned to when parsing
    """
    def __init__(self, parent, revert = True):
        self.parent = parent
        self.chk = parent.tell()
        self._revert = revert
    def revert(self):
        self.parent.seek(self.chk)
    def __enter__(self, ):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._revert:
            self.revert()

class FileStreamReaderException(IOError):
    pass

class SearchStream(metaclass=abc.ABCMeta):
    """
    Represents a stream from which we can pull block of data.
    Just provides a core interface with which we can work
    """

    @abc.abstractmethod
    def read(self, n=-1):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def readline(self):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def seek(self, *args, **kwargs):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def tell(self):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def find(self, tag, start=None, end=None):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def rfind(self, tag, start=None, end=None):
        raise NotImplementedError("SearchStream is a base class")
    @abc.abstractmethod
    def tag_size(self, tag):
        raise NotImplementedError("SearchStream is a base class")

    @abc.abstractmethod
    def __iter__(self):
        ...
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

class ByteSearchStream(SearchStream):
    """
    A stream that is implemented for searching in byte strings
    """
    def __init__(self, data, encoding="utf-8", **kw):
        """
        :param data:
        :type data: bytearray
        :param encoding:
        :type encoding:
        :param kw:
        :type kw:
        """
        self._data = data
        self._encoding = encoding
        self._kw = kw
        self._stream = None
        self._wasopen = None
    def __enter__(self):
        self._stream = io.BytesIO(self._data)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream.close()
    def __repr__(self):
        cls = type(self)
        stream = textwrap.shorten(repr(
            self._data
                if self._stream is None else
            self._stream
        ), 100)
        return f"{cls.__name__}({stream})"
    def __iter__(self):
        return iter(self._stream)
    def read(self, n=-1):
        return self._stream.read(n).decode(self._encoding)
    def readline(self):
        return self._stream.readline().decode(self._encoding)
    def seek(self, *args, **kwargs):
        return self._stream.seek(*args, **kwargs)
    def tell(self):
        return self._stream.tell()
    def encode_tag(self, tag):
        if not isinstance(tag, bytes):
            tag = tag.encode(self._encoding)
        return tag
    def find(self, tag, start=None, end=None):
        enc_tag = self.encode_tag(tag)
        if start is None:
            start = self.tell()
        if end is None:
            end = -1
        arg_vec = [enc_tag, start, end]
        return self._data.find(*arg_vec)
    def rfind(self, tag, start=None, end=None):
        enc_tag = self.encode_tag(tag)
        if start is None:
            start = 0
        if end is None:
            end = self.tell()
        arg_vec = [enc_tag, start, end]
        return self._data.rfind(*arg_vec)
    def tag_size(self, tag):
        enc_tag = self.encode_tag(tag)
        return len(enc_tag)

class FileSearchStream(SearchStream):
    """
    A stream that is implemented for searching in mmap-ed files
    """
    def __init__(self, file, mode="r", binary=True, encoding="utf-8", **kw):
        self._file = file
        if binary:
            mode = mode.strip("+b") + "+b"
        else:
            mode = mode.strip('+') + '+'
        self._mode = mode
        self._encoding = encoding
        self._kw = kw
        self._stream = None
        self._wasopen = None
    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}({self._file}, {self._mode!r})"
    def __enter__(self):
        if isinstance(self._file, str):
            self._wasopen = False
            self._fstream = open(self._file, mode=self._mode, **self._kw)
        else:
            self._wasopen = True
            self._fstream = self._file
        self._stream = mmap(self._fstream.fileno(), 0)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream.close()
        if not self._wasopen:
            self._fstream.close()
    def __iter__(self):
        new_pos = self._stream.tell()
        not_exhausted = True
        while not_exhausted:
            line = self._stream.readline()
            old_pos = new_pos
            new_pos = self._stream.tell()
            not_exhausted = new_pos > old_pos
            if not_exhausted:
                yield line
    def read(self, n=-1):
        return self._stream.read(n).decode(self._encoding)
    def readline(self):
        return self._stream.readline().decode(self._encoding)
    def seek(self, *args, **kwargs):
        return self._stream.seek(*args, **kwargs)
    def tell(self):
        return self._stream.tell()
    def find(self, tag, start=None, end=None):
        enc_tag = tag.encode(self._encoding)
        if start is None:
            start = self.tell()
        if end is None:
            end = -1
        arg_vec = [enc_tag, start, end]
        return self._stream.find(*arg_vec)
    def rfind(self, tag, start=None, end=None):
        enc_tag = tag.encode(self._encoding)
        if start is None:
            start = 0
        if end is None:
            end = self.tell()
        arg_vec = [enc_tag, start, end]

        return self._stream.rfind(*arg_vec)
    def tag_size(self, tag):
        enc_tag = tag.encode(self._encoding)
        return len(enc_tag)

class StringSearchStream(SearchStream):
    """
    A stream that is implemented for searching in strings.
    Current implementation creates a `StringIO` buffer to support `read`/`tell`/etc.
    This is very memory inefficient, but we're not winning performance awards for
    any of this anyway
    """
    def __init__(self, string):
        """
        :param string:
        :type string: str
        """
        self._data = string
        self._stream = None

    def __enter__(self):
        self._stream = io.StringIO(self._data)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream.close()
    def __iter__(self):
        return iter(self._stream)

    def read(self, n=-1):
        return self._stream.read(n)
    def readline(self):
        return self._stream.readline()
    def seek(self, *args, **kwargs):
        return self._stream.seek(*args, **kwargs)
    def tell(self):
        return self._stream.tell()
    def find(self, tag, start=None, end=None):
        if start is None:
            start = self.tell()
        if end is None:
            end = len(self._data) + 1
        arg_vec = [tag, start, end]

        return self._data.find(*arg_vec)
    def rfind(self, tag, start=None, end=None):
        if start is None:
            start = 0
        if end is None:
            end = self.tell()
        arg_vec = [tag, start, end]
        return self._data.rfind(*arg_vec)
    def tag_size(self, tag):
        return len(tag)

class SearchStreamReader:
    """
    Represents a reader which implements finding chunks of data in a stream
    """

    def __init__(self, stream):
        """
        :param stream:
        :type stream: SearchStream
        """
        self.stream = stream
    def __enter__(self):
        self.stream.__enter__()
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stream.__exit__(exc_type, exc_val, exc_tb)
    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}({self.stream})"

    def _find_tag(self, tag,
                  skip_tag=True,
                  seek=True,
                  direction='forward'
                  ):
        """
        Finds a tag in a file

        :param header: a tag specifying a header to look for + optional follow-up processing/offsets
        :type header: FileStreamerTag
        :return: position of tag
        :rtype: int
        """
        with FileStreamCheckPoint(self, revert=False) as chk:
            if direction == 'forward':
                pos = self.stream.find(tag)
            else:
                pos = self.stream.rfind(tag)
            if pos >= 0:
                if skip_tag:
                    tag_size = self.stream.tag_size(tag)
                else:
                    tag_size = 0

                if direction == 'forward':
                    pos += tag_size
                else:
                    pos -= tag_size
                if seek:
                    self.stream.seek(pos)
            elif pos == -1:
                chk.revert()
        return pos

    def find_tag(self,
                 tag,
                 skip_tag=None,
                 seek=None,
                 allow_terminal=False
                 ):
        """
        Finds a tag in a file

        :param header: a tag specifying a header to look for + optional follow-up processing/offsets
        :type header: FileStreamerTag
        :return: position of tag
        :rtype: int
        """
        if isinstance(tag, str):
            tags = FileStreamerTag(tag)
        elif isinstance(tag, dict):
            tag = tag.copy()
            t = tag['tag']
            del tag['tag']
            tags = FileStreamerTag(t, **tag)
        else:
            tags = tag

        pos = -1
        if skip_tag is None:
            skip_tag = tags.skip_tag
        if seek is None:
            seek = tags.seek
        for i, tag in enumerate(tags.tags):
            pos = self._find_tag(tag,
                                 skip_tag=skip_tag,
                                 seek=seek
                                 )
            if pos == -1:
                continue

            follow_ups = tags.skips
            if follow_ups is not None:
                for tag in follow_ups:
                    self.stream.seek(pos + 1)
                    p = self.find_tag(tag)
                    if allow_terminal or p > -1:
                        pos = p

            offset = tags.offset
            if offset is not None:
                # why are we using self._stream.tell here...?
                # I won't touch it for now but I feel like it should be pos
                pos = self.stream.tell() + offset
                self.stream.seek(pos)
            break # first match in alternatives

        return pos

    class TagSentinels(enum.Enum):
        EOF = 'end_of_file'
    def get_tagged_block(self, tag_start, tag_end, validator=None, allow_terminal=False, block_size=500):
        """
        Pulls the string between tag_start and tag_end

        :param tag_start:
        :type tag_start: FileStreamerTag or None
        :param tag_end:
        :type tag_end: FileStreamerTag
        :return:
        :rtype:
        """

        block = None
        while block is None:
            if tag_start is not None:
                    start = self.find_tag(tag_start, allow_terminal=False)
                    if start >= 0:
                        with FileStreamCheckPoint(self):
                            end = self.find_tag(tag_end, allow_terminal=allow_terminal, seek=False)
                        if end > start:
                            block = self.stream.read(end-start)
                        elif allow_terminal and end < 0:
                            block = self.stream.read()
                        else:
                            block = self.TagSentinels.EOF
                    else:
                        block = self.TagSentinels.EOF
            else:
                start = self.tell()
                with FileStreamCheckPoint(self):
                    end = self.find_tag(tag_end, allow_terminal=allow_terminal, seek=False)
                if end >= start:
                    block = self.stream.read(end-start)
                elif allow_terminal and end < 0:
                    block = self.stream.read()
                else:
                    block = self.TagSentinels.EOF

            if (
                    block is not None
                    and block is not self.TagSentinels.EOF
                    and validator is not None
                    and not validator(block)
            ):
                block = None

        if block is self.TagSentinels.EOF:
            block = None
        return block

    def parse_key_block(self,
                        tag_start=None,
                        tag_end=None,
                        mode="Single",
                        validator=None,
                        parser=None,
                        parse_mode="List",
                        num=None,
                        pass_context=False,
                        **ignore
                        ):
        """Parses a block by starting at tag_start and looking for tag_end and parsing what's in between them

        :param key: registered key pattern to pull from a file
        :type key: str
        :return:
        :rtype:
        """
        # if tag_start is None:
        #     raise FileStreamReaderException("{}.{}: needs a '{}' argument".format(
        #         type(self).__name__,
        #         "parse_key_block",
        #         "tag_start"
        #     ))
        if tag_end is None:
            raise FileStreamReaderException("{}.{}: needs a '{}' argument".format(
                type(self).__name__,
                "parse_key_block",
                "tag_end"
            ))
        if mode == "List":
            with FileStreamCheckPoint(self):
                # we do this in a checkpointed fashion only for list-type tokens
                # for all other tokens we introduce an ordering to apply when checking
                # does it need to be done like this... probably not?
                # I suppose we could be a significantly more efficient by returning a
                # generator statement in these multi-block cases
                # and then introducing a sorting order across these multi-blocks
                # that tells us which to check first, second, etc.
                # but this is probably good enough
                if isinstance(num, int):
                    blocks = [None]*num
                    if parser is None:
                        parser = lambda a, reader=None:a

                    i = 0 # protective
                    for i in range(num):
                        block = self.get_tagged_block(tag_start, tag_end, validator=validator)
                        if block is None:
                            break
                        if parse_mode != "List":
                            if pass_context:
                                block = parser(block, reader=self)
                            else:
                                block = parser(block)
                        blocks[i] = block

                    if parse_mode == "List":
                        blocks = parser(blocks[:i+1])
                else:
                    blocks = []
                    block = self.get_tagged_block(tag_start, tag_end, validator=validator)
                    if parser is None:
                        parser = lambda a, reader=None:a
                    while block is not None:
                        if parse_mode != "List":
                            if pass_context:
                                block = parser(block, reader=self)
                            else:
                                block = parser(block)

                        blocks.append(block)
                        block = self.get_tagged_block(tag_start, tag_end, validator=validator)

                    if parse_mode == "List":
                        if pass_context:
                            blocks = parser(blocks, reader=self)
                        else:
                            blocks = parser(blocks)
                return blocks
        else:
            block = self.get_tagged_block(tag_start, tag_end, validator=validator)
            if parser is not None:
                block = parser(block)
            return block

    def read(self, n=1):
        return self.stream.read(n)
    def readline(self):
        return self.stream.readline()
    def seek(self, *args, **kwargs):
        return self.stream.seek(*args, **kwargs)
    def tell(self):
        return self.stream.tell()
    def find(self, tag):
        return self.stream.find(tag)
    def rfind(self, tag):
        return self.stream.rfind(tag)
    def skip_tag(self, tag):
        return self.stream.skip_tag(tag)
    def rskip_tag(self, tag):
        return self.stream.rskip_tag(tag)

class FileStreamReader(SearchStreamReader):
    """
    Represents a file from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, file, mode="r", encoding="utf-8", **kw):
        stream = FileSearchStream(file, mode=mode, encoding=encoding, **kw)
        super().__init__(stream)
class StringStreamReader(SearchStreamReader):
    """
    Represents a string from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, string):
        stream = StringSearchStream(string)
        super().__init__(stream)
class ByteStreamReader(SearchStreamReader):
    """
    Represents a string from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, string, encoding="utf-8", **kw):
        stream = ByteSearchStream(string, encoding=encoding, **kw)
        super().__init__(stream)

class FileStreamerTag:
    def __init__(self,
                 tag_alternatives=None,
                 follow_ups=None,
                 offset=None,
                 direction="forward",
                 skip_tag=True,
                 seek=True
                 ):
        if tag_alternatives is None:
            raise FileStreamReaderException("{} needs to be supplied with some set of tag_alternatives to look for".format(
                type(self).__name__
            ))
        self.tags = (tag_alternatives,) if isinstance(tag_alternatives, str) else tag_alternatives
        self.skips = (follow_ups,) if isinstance(follow_ups, (str, FileStreamerTag)) else follow_ups
        self.offset = offset
        self.direction = direction
        self.skip_tag = skip_tag
        self.seek = seek

class LineByLineParser(metaclass=abc.ABCMeta):
    def __init__(self, stream, binary=True, encoding='utf-8', max_nesting_depth=-1, ignore_comments=False):
        """
        :param stream:
        :type stream: SearchStream
        """
        self.stream = stream
        self.binary = binary
        self.encoding = encoding
        self.max_nesting_depth = max_nesting_depth
        self.ignore_comments = ignore_comments

    def __enter__(self):
        self.stream.__enter__()
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stream.__exit__(exc_type, exc_val, exc_tb)
    @abc.abstractmethod
    def check_tag(self, line, depth:int=0, active_tag=None, label:str=None, history:list[str]=None):
        ...
    def handle_block_line(self, label, line, depth=0, history:list[str]=None):
        return line
    def handle_block(self, label, block, depth=0):
        return block
    class LineReaderTags(enum.Enum):
        RESETTING_BLOCK_END = 'implict_end'
        BLOCK_END = 'end'
        BLOCK_START = 'block'
        COMMENT = "comment"
        SKIP = "skip"
        VALUE = "value"
    def read_stream_line(self, binary=None):
        if binary is None:
            binary = self.binary
        data = next(iter(self.stream))
        if not binary and hasattr(data, 'decode'):
            data = data.decode(self.encoding)
        return data
    def stream_iter(self, binary=None):
        if binary is None:
            binary = self.binary
        line_iter = iter(self.stream)
        try:
            data = next(line_iter)
        except StopIteration:
            yield None
        test = not binary and hasattr(data, 'decode')
        if test:
            data = data.decode(self.encoding)
        yield data
        if test:
            for data in line_iter:
                data = data.decode(self.encoding)
                yield data
        else:
            for line in line_iter:
                yield line

    def find_next_block(self, binary=None, ignore_comments=None, max_nesting_depth=None,
                        aggregate_values=True, depth=0):
        if ignore_comments is None:
            ignore_comments = self.ignore_comments
        if max_nesting_depth is None:
            max_nesting_depth = self.max_nesting_depth

        stream_pos = self.stream.tell()
        block_data = []
        empty = True
        active_tag = None
        label = None
        try:
            for i,line in enumerate(self.stream_iter(binary=binary)):
                if line is None: break
                empty = False
                next_tag = self.check_tag(line, depth, active_tag=active_tag, label=label, history=block_data)
                tag_label = tag_body = None
                if not isinstance(next_tag, self.LineReaderTags) and next_tag is not None:
                    if isinstance(next_tag, (str, bytes)):
                        tag_label = next_tag
                        next_tag = None
                    elif len(next_tag) == 2:
                        next_tag, tag_body = next_tag
                    else:
                        next_tag, tag_label, tag_body = next_tag
                if next_tag == self.LineReaderTags.SKIP:
                    stream_pos = self.stream.tell()
                    continue
                if ignore_comments and next_tag == self.LineReaderTags.COMMENT:
                    stream_pos = self.stream.tell()
                    continue
                if next_tag is self.LineReaderTags.VALUE:
                    if not aggregate_values:
                        next_tag = self.LineReaderTags.BLOCK_START
                    else:
                        if tag_label is not None:
                            val = self.handle_block_line(tag_label, line, depth)
                        elif tag_body is None:
                            raise ValueError("got {} tag but no associated `(key,value)` pair")
                        else:
                            val = self.handle_block_line(label, line, depth)
                        block_data.append(val)
                        stream_pos = self.stream.tell()
                        continue

                if next_tag is self.LineReaderTags.BLOCK_END:
                    if tag_body is not None: block_data.append(tag_body)
                    stream_pos = None
                    break
                elif next_tag is self.LineReaderTags.BLOCK_START:
                    if active_tag is None:
                        label = tag_label
                        if tag_body is not None: block_data.append(tag_body)
                    else:
                        if max_nesting_depth < 0 or depth < max_nesting_depth:
                            self.stream.seek(stream_pos)
                            sub_block = self.find_next_block(binary=binary, ignore_comments=ignore_comments,
                                                             max_nesting_depth=max_nesting_depth, depth=depth+1)
                            block_data.append(sub_block)
                        else:
                            break
                elif next_tag is self.LineReaderTags.RESETTING_BLOCK_END:
                    break
                elif active_tag is not None and next_tag is not None and next_tag != active_tag:
                    if depth == 0 or (max_nesting_depth > 0 and depth >= max_nesting_depth):
                        break
                    else:
                        self.stream.seek(stream_pos)
                        sub_block = self.find_next_block(binary=binary, ignore_comments=ignore_comments,
                                                         max_nesting_depth=max_nesting_depth, depth=depth + 1)
                        block_data.append(sub_block)
                else:
                    if tag_body is not None: block_data.append(tag_body)
                    if next_tag is not None and active_tag is None:
                        active_tag = next_tag
                        stream_pos = None
                        break
                    block_data.append(self.handle_block_line(label, line, depth))
                stream_pos = self.stream.tell()
                if active_tag is None and next_tag is not None:
                    active_tag = next_tag
        except StopIteration:
            pass
        finally:
            if stream_pos is not None:
                self.stream.seek(stream_pos)

        if empty:
            return None
        else:
            if label is None:
                return self.handle_block(None, block_data)
            else:
                return {label:self.handle_block(label, block_data)}

    MAX_BLOCKS = sys.maxsize # a debug tool
    def __iter__(self):
        block = self.find_next_block()
        for i in range(self.MAX_BLOCKS):
            if block is None: break
            yield block
            block = self.find_next_block()

class FileLineByLineReader(LineByLineParser):
    """
    Represents a file from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, file,
                 mode="r", binary=False, encoding="utf-8",
                 ignore_comments=False, max_nesting_depth=-1, **kw):
        stream = FileSearchStream(file, binary=binary, mode=mode, encoding=encoding, **kw)
        super().__init__(stream,
                         binary='b' in stream._mode, encoding=encoding,
                         ignore_comments=ignore_comments, max_nesting_depth=max_nesting_depth
                         )
class StringLineByLineReader(LineByLineParser):
    """
    Represents a string from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, string, ignore_comments=False, max_nesting_depth=-1):
        stream = StringSearchStream(string)
        super().__init__(stream, binary=False, ignore_comments=ignore_comments, max_nesting_depth=max_nesting_depth)
class ByteLineByLineReader(LineByLineParser):
    """
    Represents a string from which we'll stream blocks of data by finding tags and parsing what's between them
    """
    def __init__(self, string, encoding="utf-8", ignore_comments=False, max_nesting_depth=-1, **kw):
        stream = ByteSearchStream(string, encoding=encoding, **kw)
        super().__init__(stream, binary=True, encoding=encoding,
                         ignore_comments=ignore_comments,
                         max_nesting_depth=max_nesting_depth)



import numpy as np
from .CoordinateSystemConverter import CoordinateSystemConverters as converters, CoordinateSystemConverter
from .CoordinateUtils import is_multiconfig, mc_safe_apply

__all__ = [
    "CoordinateSystem",
    "BaseCoordinateSystem",
    "CoordinateSystemError"
]

__reload_hook__ = ['.CoordinateSystemConverter', '.CoordinateUtils']

######################################################################################################
##
##                                   CoordinateSystem Class
##
######################################################################################################

class CoordinateSystem:
    """A representation of a coordinate system. It doesn't do much on its own but it *does* provide a way
    to unify internal, cartesian, derived type coordinates

    """
    def __init__(self,
                 name=None, basis=None, matrix=None, inverse=None,
                 dimension=None, origin=None, coordinate_shape=None,
                 jacobian_prep=None,
                 converter_options=None,
                 **extra
                 ):
        """
        Sets up the CoordinateSystem object

        :param name: a name to give to the coordinate system
        :type name: str
        :param basis: a basis for the coordinate system
        :type basis:
        :param matrix: an expansion coefficient matrix for the set of coordinates in its basis
        :type matrix: np.ndarray | None
        :param dimension: the dimension of a single configuration in the coordinate system (for validation)
        :type dimension: Iterable[None | int]
        :param jacobian_prep: a function for preparing coordinates to be used in computing the Jacobian
        :type jacobian_prep: function | None
        :param coordinate_shape: the actual shape of a single coordinate in the coordinate system
        :type coordinate_shape: iterable[int]
        """

        if dimension is None:
            if origin is not None:
                dimension = origin.shape
            elif matrix is not None:
                dimension = (matrix.shape[-1],)

        # if coordinate_shape is None:
        #     coordinate_shape = dimension
        # if matrix is not None and (coordinate_shape is not None and coordinate_shape[-1] != matrix.shape[-1]):
        #     raise CoordinateSystemError(
        #         "{}: expansion matrix shape {} must be compatible with coordinate shape {}".format(
        #             type(self).__name__,
        #             matrix.shape,
        #             coordinate_shape
        #         )
        #     )

        if converter_options is None:
            converter_options = {}
        converter_options = dict(extra, **converter_options)

        self.name = name
        self._basis = basis
        self._matrix = matrix
        self._inv = inverse
        self._origin = origin
        self._dimension = dimension
        self.jacobian_prep = jacobian_prep
        self.coordinate_shape = coordinate_shape
        self.converter_options = converter_options
        self._validate()

    def to_state(self, serializer=None):
        return {
            'name':self.name,
            'basis':serializer.serialize(self._basis),
            'matrix':self._matrix,
            'inverse':self._inv,
            'dimension':self._dimension,
            'origin':self._origin,
            'coordinate_shape':self.coordinate_shape,
            'jacobian_prep':self.jacobian_prep,
            'converter_options':self.converter_options
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            name=data['name'],
            basis=serializer.deserialize(data['basis']),
            matrix=data['matrix'],
            inverse=data['inverse'],
            dimension=data['dimension'],
            origin=data['origin'],
            coordinate_shape=data['coordinate_shape'],
            jacobian_prep=serializer.deserialize(data['jacobian_prep']),
            converter_options=data['converter_options']
        )

    def __call__(self, coords, **opts): # just a convenience...
        from .CoordinateSet import CoordinateSet
        opts = dict(self.converter_options, **opts)
        return CoordinateSet(coords,
                             self,
                             converter_options=opts
                             )
    def pre_convert(self, system):
        """
        A hook to allow for handlign details before converting
        :param system:
        :type system:
        :return:
        :rtype:
        """
        pass

    def _validate(self):
        if self._matrix is None:
            pass
        elif len(self._matrix.shape) != 2:
            raise CoordinateSystemError("{}: expansion matrix must be a matrix".format(type(self).__name__))
        # elif self._matrix.shape[0] != self._matrix.shape[1]:
        #     raise CoordinateSystemException("{}: expansion matrix must square".format(type(self).__name__))

        return True

    @property
    def basis(self):
        """
        :return: The basis for the representation of `matrix`
        :rtype: CoordinateSystem
        """
        return self._basis
    @property
    def origin(self):
        """
        :return: The origin for the expansion defined by `matrix`
        :rtype: np.ndarray
        """

        return self._origin
    @origin.setter
    def origin(self, orig):
        self._origin = orig
    @property
    def matrix(self):
        """
        The matrix representation in the `CoordinateSystem.basis`
        `None` is shorthand for the identity matrix

        :return: mat
        :rtype:  np.ndarray
        """
        return self._matrix
    @matrix.setter
    def matrix(self, mat):
        self._matrix = mat
    @property
    def inverse(self):
        """
        The inverse of the representation in the `basis`.
        `None` is shorthand for the inverse or pseudoinverse of `matrix`.

        :return: inv
        :rtype:  np.ndarray
        """
        if self._inv is None and self._matrix is not None:
            square_Q=self.matrix.shape[0] == self.matrix.shape[1]
            self._inv = (np.linalg.inv if square_Q else np.linalg.pinv)(self.matrix)
        return self._inv
    @inverse.setter
    def inverse(self, mat):
        self._inv = mat
    @property
    def dimension(self):
        """
        The dimension of the coordinate system.
        `None` means unspecified dimension

        :return: dim
        :rtype: int or None
        """
        return self._dimension

    def register_converter(self, system, conversion):
        converters.register_converter(self, system, conversion)

    def converter(self, system):
        """
        Gets the converter from the current system to a new system

        :param system: the target CoordinateSystem
        :type system: CoordinateSystem
        :return: converter object
        :rtype: CoordinateSystemConverter
        """

        return converters.get_converter(self, system)

    @staticmethod
    def _apply_system_matrix(basis, coords, matrix, input_coordinate_shape, target_coordinate_shape):

        odims = target_coordinate_shape
        nones = len([a for a in odims if a is None])
        if nones > 1:
            raise CoordinateSystemError(
                "Basis {} has indeterminate shape {}. Only one None is allowed".format(
                    basis,
                    odims
                ))

        shape = coords.shape
        n = matrix.shape[1]
        needs_reshape = shape[-1] != n

        if needs_reshape:
            cum_prods = np.cumprod(np.flip(shape))
            pos = np.where(cum_prods==n)[0]
            if len(pos) == 0:
                raise ValueError("Coordinate array of shape '{}' can't work with an expansion matrix of shape '{}'".format(
                    shape,
                    matrix.shape
                ))
            pos=pos[0]
            excess_shape = shape[:-(pos+1)]
            new_shape = shape[:-(pos+1)]+(n,)
            coords = np.reshape(coords, new_shape)
        else:
            pos = -1
            excess_shape = shape[:-1]

        # set up the basic expansion coordinates in the more primitive system
        coords = np.tensordot(coords, matrix, axes=((-1,), (1,)))
        # then reshape so as to actually fit the shape of the basis
        out_shape = excess_shape + target_coordinate_shape
        if nones == 1:
            the_stuff = np.prod([z for z in out_shape if z is not None])
            leftover = int(np.prod(coords.shape) / the_stuff)
            out_shape = tuple(z if z is not None else leftover for z in out_shape)
        coords = np.reshape(coords, out_shape)

        return coords

    class _convert_caller:# for multiprocessing
        def __init__(self, converter, kw, do_many):
            self.converter = converter
            self.kwargs = kw
            self.do_many = do_many
        def __call__(self, coords, *args, **kwargs):
            ops = dict(self.kwargs, **kwargs)
            if self.do_many:
                return self.converter.convert_many(coords, **ops)
            else:
                return self.converter.convert(coords, **ops)
    def convert_coords(self, coords, system, converter=None, apply_pre_converter=False, **kw):
        """
        Converts coordiantes from the current coordinate system to _system_

        :param coords:
        :type coords: CoordinateSet
        :param system:
        :type system: CoordinateSystem
        :param kw: options to be passed through to the converter object
        :type kw:
        :return: the converted coordiantes
        :rtype: tuple(np.ndarray, dict)
        """
        if apply_pre_converter:
            self.pre_convert(system)
            system.pre_convert(self)

        converter_opts = self.converter_options
        if converter_opts is None:
            converter_opts = {}
        if system is self:
            return coords, converter_opts
        sysops = system.converter_options
        if sysops is None:
            sysops = {}
        ops = dict(sysops, **converter_opts)
        kw = dict(ops, **kw)

        if self.matrix is not None:
            # This very commonly means that we're doing an expansion in some coordinate set,
            #   but there's an equilibrium value or 'origin' that we need to shift off...
            # For example, with normal modes we need to shift by the equilibrium value of the coordinates
            #   both when we convert _to_ normal modes (in which case the modes are `system`) and when we convert _from_
            #   normal modes (in which case the modes are `self`)
            coords = self._apply_system_matrix(self.basis, coords, self.matrix, self.coordinate_shape, self.basis.coordinate_shape)
            orig = self.origin
            if orig is not None:
                extra = coords.ndim-orig.ndim
                if extra>0:
                    orig = np.reshape(orig, (1,)*extra+orig.shape)
                coords = coords + orig
            return self.basis.convert_coords(coords, system, converter=converter, **kw)

        elif system.matrix is not None:
            coords, convs = self.convert_coords(coords, system.basis, converter=converter, **kw)
            inv = system.inverse
            orig = system.origin
            if orig is not None:
                extra = coords.ndim - orig.ndim
                if extra > 0:
                    orig = np.reshape(orig, (1,) * extra + orig.shape)
                coords = coords - orig
            coords = self._apply_system_matrix(system, coords,
                                               inv,
                                               system.basis.coordinate_shape,
                                               system.coordinate_shape
                                               )
            return coords, convs
        else:
            # print("> okkkay", kw['return_derivs'] if 'return_derivs' in kw else 'nooooooo')
            if converter is None:
                converter = self.converter(system)
            fun = self._convert_caller(converter, kw.copy(), is_multiconfig(coords))
            new_coords = mc_safe_apply(fun, coords=coords)
            # new_coords = fun(coords)
            # print("...wtf", kw['return_derivs'] if 'return_derivs' in kw else 'nooooooo')
            return new_coords

    def rescale(self, scaling, in_place=False):
        if not in_place:
            import copy
            new = copy.copy(self)
            return new.rescale(scaling, in_place=True)
        if self.matrix is not None:
            self.matrix = self.matrix * scaling[np.newaxis, :]
            if self._inv is not None:
                self._inv = self._inv / scaling[:, np.newaxis]
        else:
            self.matrix = np.diag(scaling)
        return self
    def rotate(self, rot, in_place=False):
        if not in_place:
            import copy
            new = copy.copy(self)
            return new.rotate(rot, in_place=True)
        if self.matrix is not None:
            self.matrix = np.dot(self.matrix, rot)
            if self._inv is not None:
                self._inv = np.dot(rot, self._inv)
        else:
            self.matrix = rot
        return self


    def displacement(self, amts):
        """
        Generates a displacement or matrix of displacements based on the vector or matrix amts
        The relevance of this method has become somewhat unclear...

        :param amts:
        :type amts: np.ndarray
        :return:
        :rtype: np.ndarray
        """
        return amts # I used to think this would be relevant... but I don't really know now if it is
        # if self.matrix is None:
        #     return amts
        # else:
        #     if isinstance(amts, (float, int, np.integer, np.float)):
        #         amts = np.full((1, self.matrix.shape[-1]), amts)
        #     else:
        #         amts = np.asarray(amts)
        #         if amts.ndim == 1:
        #             amts = np.broadcast_to(np.reshape(amts, amts.shape + (1,)), amts.shape + self.matrix.shape[-1:])
        #     return np.matmul(amts, self.matrix)

    def derivatives(self,
                    coords,
                    function,
                    order=1,
                    coordinates=None,
                    result_shape=None,
                    **finite_difference_options
                    ):
        """
        Computes derivatives for an arbitrary function with respect to this coordinate system.
        Basically a more flexible version of `jacobian`.

        :param function:
        :type function:
        :param order:
        :type order:
        :param coordinates:
        :type coordinates:
        :param finite_difference_options:
        :type finite_difference_options:
        :return: derivative tensor
        :rtype: np.ndarray
        """

        from McUtils.Zachary import FiniteDifferenceDerivative

        deriv_tensor = None
        if order > 0:
            self_shape = self.coordinate_shape
            if self_shape is None:
                self_shape = coords.shape[1:]
            if self_shape is None:
                raise CoordinateSystemError(
                    "{}.{}: 'coordinate_shape' {} must be tuple of ints".format(
                        type(self).__name__,
                        'derivatives',
                        self_shape
                    ))

            # if other_shape is None:
            #     raise CoordinateSystemException(
            #         "{}.{}: 'coordinate_shape' {} must be tuple of ints".format(
            #             type(self).__name__,
            #             'jacobian',
            #             other_shape
            #         ))

            for k, v in zip(
                    ('mesh_spacing',),
                    (.001,)
            ):
                if k not in finite_difference_options:
                    finite_difference_options[k] = v

            deriv = FiniteDifferenceDerivative(
                function,
                function_shape=(self_shape, result_shape),
                **finite_difference_options
            )(coords)

            deriv_tensor = deriv.derivative_tensor(order, coordinates=coordinates)

        if deriv_tensor is None:
            raise CoordinateSystemError("derivative order {} <= 0".format(order))

        return deriv_tensor

    class _converter: # for multiprocessing
        def __init__(self, system, deriv_key, parent, num_derivs, convert_kwargs):
            self.system = system
            self.deriv_key = deriv_key
            self.parent = parent
            self.num_derivs = num_derivs
            self.convert_kwargs = convert_kwargs

        def __call__(self, c, *args, **kwargs):
            if self.num_derivs is None:
                return self.parent.convert_coords(c, self.system,  apply_pre_converter=False, **self.convert_kwargs)[0]
            else:
                parent = self.parent
                s = self.system
                num_derivs = self.num_derivs
                dk = self.deriv_key
                convert_kwargs = self.convert_kwargs
                crds, opts = parent.convert_coords(c, s, return_derivs=num_derivs, apply_pre_converter=False, **convert_kwargs)
                # we now have to reshape the derivatives because mc_safe_apply is only applied to the coords -_-
                # should really make that function manage deriv shapes too, but don't know how to _tell_ it that I
                # want it to
                derivs = opts[dk]
                # we also want to only do the derivatives on the highest-order analytical
                # derivative that we have
                if isinstance(derivs, np.ndarray):  # just protection for the next step, basically if we only get firsts out
                    derivs = [derivs]
                derivs = derivs[num_derivs - 1]  # so that we can set the derivative order below the total possible returned...

                # now we figure out how much shape is in 'c'
                c_dims = np.prod(c.shape)
                # and we figure out how much of the derivs to toss out to account for it
                d_dims = np.cumprod(derivs.shape)
                pos = np.where(d_dims == c_dims)[0]
                if len(pos) == 0:
                    raise ValueError(
                        "Shape mismatch in Jacobian (coordinates with shape {} returned derivatives with shape {})".format(
                            c.shape,
                            derivs.shape
                        ))

                new_deriv_shape = c.shape + derivs.shape[pos[0] + 1:]
                derivs = derivs.reshape(new_deriv_shape)

                return derivs

    return_derivs_key = 'return_derivs'
    deriv_key = 'derivs'
    def jacobian(self,
                 coords,
                 system,
                 order=1,
                 coordinates=None,
                 converter_options=None,
                 all_numerical=False,
                 analytic_deriv_order=None,
                 **finite_difference_options
                 ):
        """
        Computes the Jacobian between the current coordinate system and a target coordinate system

        :param system: the target CoordinateSystem
        :type system: CoordinateSystem
        :param order: the order of the Jacobian to compute, 1 for a standard, 2 for the Hessian, etc.
        :type order: int | Iterable[int]
        :param coordinates: a spec of which coordinates to generate derivatives for (None means all)
        :type coordinates: None | iterable[iterable[int] | None
        :param mesh_spacing: the spacing to use when displacing
        :type mesh_spacing: float | np.ndarray
        :param prep: a function for pre-validating the generated coordinate values and grids
        :type prep: None | function
        :param fd_options: options to be passed straight through to FiniteDifferenceFunction
        :type fd_options:
        :return: derivative tensor
        :rtype: np.ndarray
        """

        # print(system)
        from McUtils.Zachary import FiniteDifferenceDerivative

        self.pre_convert(system)
        system.pre_convert(self)

        if converter_options is None:
            converter_options = {} # convert_coords tracks the other conversion options for us

        deriv_tensors = None # default return value

        self_shape = self.coordinate_shape
        if self_shape is None:
            self_shape = coords.shape[1:]
        if self_shape is None:
            raise CoordinateSystemError(
                "{}.{}: 'coordinate_shape' {} must be tuple of ints".format(
                    type(self).__name__,
                    'jacobian',
                    self_shape
                ))

        all_numerical = all_numerical or analytic_deriv_order == 0
        if not all_numerical:
            # sort of a hack right now: if the conversion returns 'derivs', then we use those for the FD
            # unless we got enough analytic derivatives to not need to do any more FD
            ret_d_key = self.return_derivs_key
            rd = converter_options.get(ret_d_key)
            converter_options[ret_d_key] = order if analytic_deriv_order is None else (analytic_deriv_order != 0)
            test_crd, test_opts = self.convert_coords(coords, system, order=order,
                                                      apply_pre_converter=False, **converter_options)
            if rd is None:
                del converter_options[ret_d_key]
            else:
                converter_options[ret_d_key] = rd
            deriv_key = self.deriv_key
            deriv_tensors = test_opts.get(deriv_key)
            if deriv_tensors is not None:
                # we now check to see how many derivs we got
                # so that we can decrement the order by that amount
                # if we just get back one numpy array of derivatives
                # we assume that we got back the analytic first derivatives
                if isinstance(deriv_tensors, np.ndarray):
                    deriv_tensors = [deriv_tensors]
                if analytic_deriv_order is not None:
                    if analytic_deriv_order < 0:
                        analytic_deriv_order = len(deriv_tensors)
                    num_derivs = min(len(deriv_tensors), analytic_deriv_order)
                    deriv_tensors = deriv_tensors[:num_derivs]
                else:
                    num_derivs = len(deriv_tensors)

                if isinstance(order, int):
                    if order > num_derivs:
                        deriv_tensors = [deriv_tensors[-1]]
                    else:
                        deriv_tensors = [deriv_tensors[order - 1]]
                    order = order-num_derivs
                else:
                    order = [o-num_derivs for o in order if o > num_derivs]


                kw = converter_options.copy()
                kw["order"] = order
                if ret_d_key in kw:
                    del kw[ret_d_key]

                convert = self._converter(system, deriv_key, self, num_derivs, kw)
            else:
                convert = self._converter(system, deriv_key, self, None, converter_options)
        else:
            convert = self._converter(system, None, self, None, converter_options)
            # convert = lambda c, s=system, kw=converter_options: self.convert_coords(c, s, **kw)[0]
        need_derivs = (len(order) > 0 and max(order) > 0) if not isinstance(order, int) else order > 0
        if need_derivs:
            other_shape = system.coordinate_shape
            # if other_shape is None:
            #     raise CoordinateSystemException(
            #         "{}.{}: 'coordinate_shape' {} must be tuple of ints".format(
            #             type(self).__name__,
            #             'jacobian',
            #             other_shape
            #         ))

            # set default options
            for k, v in zip(
                    ('mesh_spacing', 'prep', 'stencil'),
                    (.001, system.jacobian_prep, 7)
            ):
                if k not in finite_difference_options or finite_difference_options[k] is None:
                    finite_difference_options[k] = v

            # print("??", other_shape, coords.shape)
            # print("?>", convert(coords).shape)
            # wat = convert(coords)
            # base_shape = wat.shape[len(coords.shape) - len(self_shape):]
            # print("???", wat.shape, coords.shape, base_shape)
            deriv = FiniteDifferenceDerivative(
                convert,
                function_shape=(self_shape, None),
                **finite_difference_options
            )(coords)

            if not isinstance(order, int):
                # print(order)
                if 0 in order:
                    if order[0] != 0:
                        raise NotImplementedError("I don't want to mess with ordering so just put the 1 or 0 first...")
                    else:
                        ordo = order[1:]
                else:
                    ordo = order
            else:
                ordo = [order]
            # print(ordo)
            derivs = deriv.derivative_tensor(ordo, coordinates=coordinates)
            if isinstance(order, int):
                deriv_tensors = [derivs[0]]
            elif deriv_tensors is None:
                deriv_tensors = derivs
            else:
                deriv_tensors = deriv_tensors + derivs # currently assuming you'd put

        if isinstance(order, int):
            deriv_tensors = deriv_tensors[0]

        if deriv_tensors is None:
            raise CoordinateSystemError("derivative order '{}' less than 0".format(order))

        # print([(o, x.shape) for o, x in (zip([order], [deriv_tensors]) if isinstance(order, int) else zip(order, deriv_tensors))])
        return deriv_tensors

    def __repr__(self):
        """
        Provides a clean representation of a `CoordinateSystem` for printing
        :return:
        :rtype: str
        """
        return "CoordinateSystem({}, dimension={}, matrix={})".format(self.name, self.dimension, self.matrix)
    @classmethod
    def is_compatible(cls, self, system):
        return (
                self is system
                or self.name == system.name
                # or (isinstance(system, type) and isinstance(self, system))
        )
        # or system1.name == key_pair[0].name and system2.name == key_pair[1].name
    def has_conversion(self, system): # to be overloaded
        return False

######################################################################################################
##
##                                   CoordinateSystemException Class
##
######################################################################################################
class CoordinateSystemError(Exception):
    """
    An exception that happens inside a `CoordinateSystem` method
    """
    pass

######################################################################################################
##
##                                   BaseCoordinateSystem Class
##
######################################################################################################

class BaseCoordinateSystem(CoordinateSystem):
    """
    A CoordinateSystem object that can't be reduced further.
    A common choice might be Cartesian coordinates or internal coordinates.
    This allows us to define flexible `CoordinateSystem` subclasses that we _don't_ expect to be used as a base
    """

    def __init__(self, name, dimension=None, coordinate_shape=None, converter_options=None):
        super().__init__(name=name,
                         dimension=dimension, basis=self, coordinate_shape=coordinate_shape,
                         converter_options=converter_options
                         )

    def to_state(self, serializer=None):
        return {
            'name':self.name,
            'dimension':self._dimension,
            'coordinate_shape':self.coordinate_shape,
            'converter_options':self.converter_options
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            name=data['name'],
            dimension=data['dimension'],
            coordinate_shape=data['coordinate_shape'],
            converter_options=data['converter_options']
        )



import itertools, collections
import typing

import scipy
import scipy.sparse as sparse, numpy as np
from .. import Numputils as nput
from .. import Iterators as itut
from . import Trees as tree

__all__ = [
    "EdgeGraph",
    "MoleculeEdgeGraph"
]

class EdgeGraph:
    map: dict[int, set[int]]
    __slots__ = ["labels", "edges", "graph", "map", "_rings", "_sp_data"]
    def __init__(self, labels, edges, graph=None, edge_map=None):
        self.labels = labels
        self.edges = np.asanyarray(edges)
        if graph is None:
            graph = self.adj_mat(len(labels), self.edges)
        self.graph = graph
        if edge_map is None:
            edge_map = self.build_edge_map(self.edges)
        self.map = edge_map
        self._rings = None
        self._sp_data = None

    def graph_difference(self, other):
        diff = (other.graph - self.graph)
        if isinstance(diff, np.ndarray):
            nzr, nzc = np.nonzero(diff)
            new = diff[nzr, nzc].asarray() > 0
        else:
            nzr, nzc, vals = sparse.find(diff)
            new = vals > 0
        old = np.logical_not(new)

        return np.array([nzr[new], nzc[new]]).T, np.array([nzr[old], nzc[old]]).T

    @classmethod
    def get_edge_graph(cls, spec, num_nodes=None):
        if isinstance(spec, dict):
            edge_list = np.array([
                (i, j)
                for i in spec
                for j in spec[i]
            ])
            if num_nodes is None:
                num_nodes = len(np.unique(edge_list.flatten()))
            return cls.adj_mat(num_nodes, edge_list)
        elif isinstance(spec, (sparse.csr_matrix, sparse.coo_matrix, sparse.csc_matrix)): #TODO: better checks
            return spec
        else:
            spec = np.array(spec)
            num_nodes = len(np.unique(spec.flatten()))
            return cls.adj_mat(num_nodes, spec)

    @classmethod
    def get_edge_list(cls, spec):
        if isinstance(spec, dict):
            return [
                (i, j)
                for i in spec
                for j in spec[i]
            ]
        elif isinstance(spec, (sparse.csr_matrix, sparse.coo_matrix, sparse.csc_matrix)): #TODO: better checks
            r, c = spec.nonzero()
            return np.array([r, c]).T
        else:
            return spec

    @classmethod
    def get_edge_map(cls, spec):
        if isinstance(spec, dict):
            return spec
        elif isinstance(spec, (sparse.csr_matrix, sparse.coo_matrix, sparse.csc_matrix)):  # TODO: better checks
            return cls.get_edge_map(cls.get_edge_list(spec))
        else:
            return cls.build_edge_map(spec)

    @classmethod
    def adj_mat(cls, num_nodes, edges):
        adj = np.zeros((num_nodes, num_nodes), dtype=int)
        if len(edges) > 0:
            rows,cols = edges.T
            adj[rows, cols] = 1
            adj[cols, rows] = 1

        return sparse.csr_matrix(adj)

    def get_distances(self, indices=None):
        return sparse.csgraph.shortest_path(
            self.graph,
            directed=False,
            unweighted=True,
            indices=indices
        )

    @classmethod
    def build_edge_map(cls, edge_list, num_nodes=None):
        map = {}
        for e1,e2 in edge_list:
            if e1 not in map: map[e1] = set()
            map[e1].add(e2)
            if e2 not in map: map[e2] = set()
            map[e2].add(e1)
        if num_nodes is not None:
            for i in range(num_nodes):
                if i not in map: map[i] = set()
        return map

    @classmethod
    def _remap(cls, labels, pos, rows, cols):
        if len(rows) == 0:
            edge_list = np.array([], dtype=int).reshape(-1, 2)
        else:
            new_mapping = np.zeros(len(labels), dtype=int)
            new_mapping[pos,] = np.arange(len(pos))
            new_row = new_mapping[rows,]
            new_col = new_mapping[cols,]
            edge_list = np.array([new_row, new_col]).T

        return [labels[p] for p in pos], edge_list
    @classmethod
    def _take(cls, pos, labels, adj_mat:sparse.compressed) -> 'typing.Self':
        rows, cols, _ = sparse.find(adj_mat)
        utri = cols >= rows
        rows = rows[utri]
        cols = cols[utri]
        row_cont, _, _ = nput.contained(rows, pos)
        col_cont, _, _ = nput.contained(cols, pos)
        cont = np.logical_and(row_cont, col_cont)

        labels, edge_list = cls._remap(labels, pos, rows[cont], cols[cont])
        return cls(labels, edge_list)

    def take(self, pos):
        return self._take(pos, self.labels, self.graph)

    def split(self, backbone_pos):
        new_adj = self.graph.copy()
        for n in backbone_pos:
            for i,j in self.map[n]:
                new_adj[i,j] = 0
        ncomp, labels = sparse.csgraph.connected_components(new_adj, directed=False, return_labels=True)
        groups, _ = nput.group_by(np.arange(len(labels)), labels)
        return [
            self._take(pos, self.labels, new_adj)
            for _, pos in groups
        ]

    @classmethod
    def _subgraph_match(cls,
                        root1, labels1, edge_map1,
                        root2, labels2, edge_map2,
                        visited=None
                        ):
        if labels1[root1] != labels2[root2]:
            return False
        if len(edge_map1.get(root1, [])) != len(edge_map2.get(root2, [])):
            return False

        if visited is None:
            visited = (set(), set())
        visit1, visit2 = visited

        # queue1 = collections.deque([root1])
        tests1 = set(edge_map1.get(root1, [])) - visit1
        tests2 = set(edge_map2.get(root2, [])) - visit2
        for r1 in tests1:
            visit1.add(r1)
            for r2 in tests2:
                visit2.add(r2)
                # check if all the subgraphs match via DFS
                # we could do this without recursion, but it'd be
                # more annoying
                if cls._subgraph_match(
                    r1, labels1, edge_map1,
                    r2, labels2, edge_map2,
                    visited=(visit1, visit2)
                ):
                    tests2.remove(r2)
                    break
                else:
                    visit2.remove(r2)

        return len(tests2) == 0

    @classmethod
    def graph_match(cls, graph1:'EdgeGraph', graph2:'EdgeGraph'):
        # we do some quick prunes
        atoms1 = graph1.labels
        atoms2 = graph2.labels
        if (
                len(atoms1) != len(atoms2)
                or atoms1[0] != atoms2[0]
                or len(graph1.edges) != len(graph2.edges)
                or list(sorted(atoms1)) != list(sorted(atoms2))
                or list(sorted(len(v) for v in graph1.map.values())) != list(sorted(len(v) for v in graph2.map.values()))
        ):
            return False

        return cls._subgraph_match(
            0, graph1.labels, graph1.map,
            0, graph2.labels, graph2.map
        )

    def __eq__(self, other):
        return self.graph_match(self, other)

    @classmethod
    def build_neighborhood_graph(cls, node, labels, edge_map, ignored=None, num=1):
        edges = []
        visited = set()
        if ignored is None: ignored = []
        ignored = set(ignored)
        queue = [node]
        for i in range(num):
            new_queue = []
            for node in queue:
                visited.add(node)
                new_nodes = set(edge_map[node]) - visited - ignored
                edges.extend((node, e) for e in new_nodes)
                new_queue.extend(new_nodes)
            queue = new_queue

        edges = np.array(edges, dtype=int)
        if len(edges) == 0:
            edges = np.reshape(edges, (-1, 2))
        labels, edges = cls._remap(labels, list(visited), edges[:, 0], edges[:, 1])
        return cls(labels, edges)

    def neighbor_graph(self, root, ignored=None, num=1):
        return self.build_neighborhood_graph(root, self.labels, self.map, ignored=ignored, num=num)

    @property
    def rings(self):
        if self._rings is None:
            self._rings = self.get_rings()
        return self._rings

    @classmethod
    def find_rings_in_graph(cls, n_inds, edge_map):
        test_rings = []
        for n in range(n_inds):
            if len(edge_map[n]) < 2: continue
            visited = set()
            test_rings.append(
                tree.tree_traversal(
                    edge_map,
                    lambda parent, child, visited: (
                        visited
                        if len(visited & edge_map[child]) > 1 else
                        None
                    ),
                    root=n,
                    get_children=lambda h: [c for c in edge_map[h] if len(edge_map[c]) > 1],
                    get_item=lambda _, h: h,
                    visited=visited,
                    call_order='post',
                    traversal_ordering='bfs'
                )
            )

        filt_rings = [t for t in test_rings if t is not None]

        rings = []
        keys = set()
        for i, r1 in enumerate(filt_rings):
            for r2 in filt_rings[i:]:
                int_ring = r1 & r2
                key = tuple(sorted(int_ring))
                if key in keys: continue
                keys.add(key)
                r = cls.check_ring_in_graph(int_ring, edge_map)
                if r: rings.append(r)
        return rings

    @classmethod
    def check_ring_in_graph(cls, ring_atoms, edge_map):
        if len(ring_atoms) < 3: return False
        ring_atoms = set(ring_atoms)
        if any(len(edge_map[r] & ring_atoms) == 1 for r in ring_atoms): return False
        visited = set()
        root = next(iter(ring_atoms))
        ring = []
        return tree.tree_traversal(
                    edge_map,
                    lambda parent, child, visited: (
                        ring + [child]
                            if (parent != root and root in edge_map[child]) else
                        ring.append(child)
                    ),
                    root=root,
                    get_children=lambda h: edge_map[h] & ring_atoms,
                    get_item=lambda _, h: h,
                    visited=visited,
                    call_order='post',
                    traversal_ordering='dfs'
                )

    def get_rings(self):
        return self.find_rings_in_graph(len(self.labels), self.map)

    @classmethod
    def get_shortest_path_data(cls, graph):
        return sparse.csgraph.shortest_path(
            graph, directed=False, return_predecessors=True
        )
    @property
    def shortest_path_data(self):
        if self._sp_data is None:
            self._sp_data = self.get_shortest_path_data(self.graph)
        return self._sp_data

    @classmethod
    def get_path_from_data(cls, start, end, sp_data):
        dist_matrix, predecessors = sp_data
        path = [end]
        # straight from scipy docs...I had a better imp. at one point but dunno where it is
        for _ in range(len(dist_matrix)):  # max number of steps
            cur_node = int(predecessors[start][path[-1]])
            if cur_node < 0:
                break
            path.append(cur_node)
        return tuple(reversed(path))
    @classmethod
    def get_longest_path_from_data(cls, shortest_path_data, root=None):
        dist_matrix, predecessors = shortest_path_data
        if root is None:
            max_cols = np.argmax(dist_matrix, axis=1)
            max_row = np.argmax(dist_matrix[np.arange(len(max_cols)), max_cols])
            return cls.get_path_from_data(max_row, max_cols[max_row], shortest_path_data)
        else:
            max_col = np.argmax(dist_matrix[root], axis=0)
            return cls.get_path_from_data(root, max_col, shortest_path_data)
    def get_path(self, start, end):
        return self.get_path_from_data(start, end, self.shortest_path_data)

    @classmethod
    def compute_edge_centralities(self, indices, map):
        if nput.is_int(indices):
            return len(map[indices])
        else:
            return np.array([len(map[i]) for i in indices])

    @classmethod
    def find_longest_chain_from_breakpoints(cls,
                                            map,
                                            graph=None,
                                            rings=None,
                                            root=None,
                                            use_highest_valencies=True,
                                            shortest_path_data=None
                                            ):
        if rings is None:
            rings = cls.find_rings_in_graph(len(map), map)
        if graph is None:
            graph = cls.get_edge_graph(map)
        if len(rings) == 0:
            if shortest_path_data is None:
                shortest_path_data = cls.get_shortest_path_data(graph)
            return cls.get_longest_path_from_data(shortest_path_data, root=root)
        else:
            # in principle we have to check every set of breakpoints...
            # but we can make this more efficient by noting that a break
            # has to occur near the highest-valency position for fused rings
            # and if breaking can lead to a long ring will again be near a high-valency
            # atom in non-fused rings
            # each potential high-valency bond can be broken
            if use_highest_valencies:
                ring_valencies = [cls.compute_edge_centralities(r, map) for r in rings]
                break_pos = [
                    np.array(r)[np.where(v == np.max(v))]
                    for r,v in zip(rings, ring_valencies)
                ]
            else:
                break_pos = rings

            break_bonds = [
                [
                    (p, j)  # enumerate all ring bonds starting at the breakpos, could be more efficient
                    for p in bp
                    for j in map[p]
                    if j in r
                ]
                for r, bp in zip(rings, break_pos)
            ]
            if root is not None:
                # sort break bonds to priortize breaks by the root
                break_bonds = [
                    list(sorted(bl, key=lambda p: p[0] != root and p[1] != 1))
                    for bl in break_bonds
                ]

            # chain = None
            n_inds = len(map)
            for p_list in itertools.product(*break_bonds):
                if len({tuple(sorted(p)) for p in p_list}) < len(p_list): continue # dupe bond in fused ring
                new_map: dict[int, set[int]] = {k: v.copy() for k, v in map.items()}
                for center, bond in p_list:
                    new_map[center].remove(bond)
                    new_map[bond].remove(center)

                new_rings = cls.find_rings_in_graph(n_inds, new_map)
                if len(new_rings) == 0:
                    chain = cls.find_longest_chain_from_breakpoints(
                        new_map,
                        root=root
                    )
                    break
            else:
                raise ValueError("couldn't find a way to break rings?")

            return chain

    def find_longest_chain(self,
                           rings=None,
                           use_highest_valencies=True,
                           ):
        if rings is None:
            rings = self.rings

        return self.find_longest_chain_from_breakpoints(
            self.map,
            graph=self.graph,
            rings=self.rings,
            use_highest_valencies=use_highest_valencies,
            shortest_path_data=self.shortest_path_data
        )

    def segment_by_chains(self,
                          rings=None,
                          root=None,
                          use_highest_valencies=True,
                          ):
        return self.segment_graph_by_chains(
            self.map,
            root=root,
            graph=self.graph,
            rings=self.rings,
            use_highest_valencies=use_highest_valencies,
            shortest_path_data=self.shortest_path_data
        )

    @classmethod
    def find_graph_centroid(cls, graph, shortest_path_data=None):
        # we define the graph centroid as the point with the minimum maximum graph
        # distance over a fully-connected graph

        if shortest_path_data is None:
            shortest_path_data = cls.get_shortest_path_data(graph)
        distance_matrix, _ = shortest_path_data
        return np.argmin(np.argmax(distance_matrix, axis=1), axis=0)

    def get_centroid(self, check_fragments=True):
        if check_fragments:
            fragments = self.get_fragments()
            if len(fragments) > 0:
                centroids = []
                for f in fragments:
                    g = self.take(f)
                    centroids.append(g.get_centroid(check_fragments=False))
                return fragments, centroids

        return self.find_graph_centroid(self.graph, shortest_path_data=self.shortest_path_data)

    @classmethod
    def get_graph_fragment_indices(cls, graph):
        _, labels = sparse.csgraph.connected_components(graph, directed=False, return_labels=True)
        _, groups = nput.group_by(np.arange(len(labels)), labels)[0]
        return groups

    def get_fragments(self):
        return self.get_graph_fragment_indices(self.graph)

    @classmethod
    def _reindex_segmentss(cls, frags, remapping):
        if nput.is_int(frags[0]):
            return tuple(remapping[j] for j in frags)
        else:
            return tuple(
                cls._reindex_segmentss(f, remapping)
                for f in frags
            )
    @classmethod
    def segment_graph_by_chains(cls,
                                map: dict[int, set[int]],
                                graph: ('sparse.coo_matrix|sparse.csr_matrix|sparse.csc_matrix') = None,
                                rings=None,
                                root=None,
                                use_highest_valencies=True,
                                shortest_path_data=None
                                ):

        if len(map) == 1:
            return tuple(map.keys())

        if graph is None:
            graph = cls.get_edge_graph(map)

        segments = []

        backbone = cls.find_longest_chain_from_breakpoints(
            map,
            graph=graph,
            rings=rings,
            root=root,
            use_highest_valencies=use_highest_valencies,
            shortest_path_data=shortest_path_data
        )

        segments.append(backbone)

        new_map = {k:v.copy() for k,v in map.items()}
        for i in backbone:
            for j in new_map.get(i, set()):
                new_map[j].remove(i)
            del new_map[i]

        rem = np.delete(np.arange(graph.shape[0]), backbone)
        base_remapping = {k: i for i, k in enumerate(rem)}
        base_inv = {i: k for i, k in enumerate(rem)}
        new_map = {
            base_remapping[i]: {base_remapping[k] for k in v}
            for i,v in new_map.items()
        }

        fragments = cls.get_graph_fragment_indices(
            cls.get_edge_graph(new_map, num_nodes=len(rem))
        )

        for frag in fragments:
            if len(frag) == 1:
                return [frag]
            else:
                remapping = {k:i for i,k in enumerate(frag)}
                inverse_mapping = {i:base_inv[k] for i,k in enumerate(frag)}
                subbones = cls.segment_graph_by_chains(
                    {
                        remapping[k]: {remapping[j] for j in new_map[k]}
                        for k in frag
                    },
                    use_highest_valencies=use_highest_valencies
                )
                segments.append(
                    cls._reindex_segmentss(subbones, inverse_mapping)
                )


        return segments


    @classmethod
    def get_maximum_overlap_permutation(cls, graph_1:'EdgeGraph', graph_2:'EdgeGraph'):
        syms_1 = graph_1.labels
        syms_2 = graph_2.labels

        if any(s_1 != s_2 for s_1, s_2 in zip(syms_1, syms_2)):
            if len(itut.dict_diff(itut.counts(syms_1), itut.counts(syms_2))) > 0:
                raise ValueError(f"graph labels must agree: {syms_1} != {syms_2}")
            ordering_1 = list(sorted(range(len(syms_1)), key=syms_1.__getitem__))
            ordering_2 = list(sorted(range(len(syms_2)), key=syms_2.__getitem__))
            perm_0 = np.array(ordering_2)[np.argsort(ordering_1)]
            graph_2 = graph_2.take(perm_0)
        else:
            perm_0 = None

        bond_set_1 = {tuple(sorted(e)) for e in graph_1.edges}
        bond_set_2 = {tuple(sorted(e)) for e in graph_2.edges}

        # initial bond difference
        test_bonds = np.unique(np.array(
            list(bond_set_1 - bond_set_2)
            + list(bond_set_2 - bond_set_1)
        ))

        # permutable groups
        sym_splits, _ = nput.group_by(np.arange(len(syms_1)), np.array([ord(s) for s in syms_1]))
        perm_blocks = []
        perm_atoms = []
        for _, atom_inds in zip(*sym_splits):
            atom_inds = nput.intersection(atom_inds, test_bonds)[0]  # only permute things in the original core
            if len(atom_inds) > 0:
                perm_atoms.append(atom_inds)
                perm_blocks.append(itertools.permutations(atom_inds))

        nsym = len(syms_1)
        core_size = len(test_bonds)
        perm = np.arange(nsym)
        for full_perm in itertools.product(*perm_blocks):
            reindexing = np.arange(nsym)
            for atom_inds, new_idx in zip(perm_atoms, full_perm):
                reindexing[atom_inds,] = new_idx
            new_bond_set_1 = {
                (reindexing[i], reindexing[j])
                for (i, j) in bond_set_1
            }
            new_core = np.unique(np.array(
                list(new_bond_set_1 - bond_set_2)
                + list(bond_set_2 - new_bond_set_1)
            ))
            if len(new_core) < core_size:
                perm = reindexing

        if perm_0 is not None:
            return perm_0[perm]
        else:
            return perm

    def get_reindexing(self, other_graph):
        return self.get_maximum_overlap_permutation(other_graph, self)
    def align_labels(self, other_graph):
        perm = self.get_reindexing(other_graph)
        return self.take(perm)



class MoleculeEdgeGraph(EdgeGraph):

    def get_rings(self):
        # use rdkit's cycles...not assured to be the smallest set ;_;
        from ..ExternalPrograms.RDKit import RDMolecule
        return RDMolecule.from_coords(
            ["C"]*len(self.labels),
            coords=np.zeros((len(self.labels), 3)),
            bonds=[[int(i), int(j), 1] for i,j in self.edges]
        ).rings

    @classmethod
    def _match_motif(cls, label, neighbors, motif_root, *motif_branches):
        branch_members = set()
        for b in motif_branches:
            branch_members.update(b)
        if label not in branch_members:
            return False
        if label != motif_root:
            for b in motif_branches:
                ...
                # if l
    @classmethod
    def _idenfity_motifs(cls, label, neighbors, index=None, graph=None):
        ...
    @classmethod
    def _make_label(cls, label, neighbors, index=None, graph=None):
        if len(neighbors) == 0: return label
        if isinstance(neighbors[0], str):
            if label == "H":
                if len(neighbors) > 1:
                    return label + cls._format_atom_counts(neighbors)
                else:
                    return neighbors[0] + label
            elif label == "C":
                non_c_neighbors = [l for l in neighbors if l != "C"]
                if len(non_c_neighbors) > 0:
                    neighbors = non_c_neighbors
                elif len(neighbors) > 0:
                    return "C"+str(len(neighbors)+1)
                return label + cls._format_atom_counts(neighbors)
            else:
                non_c_neighbors = [l for l in neighbors if l != "C"]
                if len(non_c_neighbors) > 0:
                    neighbors = non_c_neighbors
                return label + cls._format_atom_counts(neighbors)
        else:
            raise ValueError("can't make atom type for neighbor depth > 2")

    def _collect_neighbor_list(self, root, depth=1, visited=None):
        if visited is None: visited = {root}
        nl = []
        ni = []
        for m in self.map.get(root, []):
            if m in visited: continue
            visited.add(m)
            if depth > 1:
                sublist, subinds = self._collect_neighbor_list(m, depth=depth-1, visited=visited)
                nl.append((self.labels[m], sublist))
                ni.append((m, subinds))
            else:
                nl.append(self.labels[m])
                ni.append(m)
        return tuple(nl), tuple(ni)


    chemical_order = ['C', 'O', 'N', 'H', 'F', 'Cl', 'Br', 'I']
    @classmethod
    def _sort_atom_types(cls, counts, chemical_order=None):
        if chemical_order is None:
            if not isinstance(cls.chemical_order, dict):
                cls.chemical_order = {
                    k:n for n,k in enumerate(cls.chemical_order)
                }
            chemical_order = cls.chemical_order
        elif not hasattr(chemical_order, 'get'):
            chemical_order = {
                k: n for n, k in enumerate(chemical_order)
            }

        if hasattr(counts, "items"):
            return sorted(
                counts.items(),
                key=lambda kv: chemical_order.get(kv[0], len(chemical_order))
            )
        else:
            return sorted(
                counts,
                key=lambda kv: chemical_order.get(kv, len(chemical_order))
            )

    @classmethod
    def _format_atom_counts(cls, neighbors):
        if isinstance(neighbors, dict):
            nbs = neighbors
        else:
            nbs = itut.counts(neighbors)
        return "".join(l+("" if k == 1 else str(k)) for l,k in cls._sort_atom_types(nbs))
    @classmethod
    def _format_ring_counts(cls, neighbors):
        return "[" + cls._format_atom_counts(neighbors) + "]"

    @classmethod
    def _atomlist_match(cls, list_1, list_2):
        count1 = itut.counts(list_1)
        count2 = itut.counts(list_2)
        for g,c in count2.items():
            if g == "_":
                continue
            else:
                c2 = count1.get(g, -1)
                if c2 > c:
                    count1[g] -= c
                    continue
                elif c2 > 0:
                    c -= c2
                    del count1[g]
        return count2.get('_', -1) < sum(count1.values())

    @classmethod
    def _bonding_pattern_matcher(cls, neighbor_lists):
        group_counts = itut.counts(neighbor_lists)
        primary_counts = itut.counts(nl[0] for nl in neighbor_lists)
        def match(test_counts, primary_counts=primary_counts, groups_counts=group_counts):
            if hasattr(test_counts, 'items'): test_counts = test_counts.items()
            unmatched = []
            primary_counts = primary_counts.copy()
            groups_counts = groups_counts.copy()
            # we handle wild cards by keeping track of the excess/unmatched patterns
            for group, count in test_counts:
                if isinstance(group, str):
                    if group == "_":
                        unmatched.append([group, count])
                        continue
                    cur_count = primary_counts.get(group, -1)
                    if cur_count < count:
                        return False
                    elif cur_count > count:
                        primary_counts[group] -= count
                    else:
                        del primary_counts[group]
                else:
                    root, rem = group
                    if root == "_" or any("_" in pat for pat in rem):
                        unmatched.append([group, count])
                        continue
                    cur_count = group_counts.get(group, -1)
                    if cur_count < count:
                        return False
                    elif cur_count > count:
                        groups_counts[group] -= count
                    else:
                        del groups_counts[group]
            # now with what we have left over, we need to match each
            # remaining "wildcard" pattern, in principle we need to do this
            # by taking ever possible combination of rems...but we'll be lazy and
            # do it greedily assuming only a few wild cards
            total_strs = sum(primary_counts.values())
            new_umatched = []
            for group, count in unmatched:
                if isinstance(group, str): #"_"
                    if total_strs >= count:
                        total_strs -= count
                    else:
                        return False
                elif group[0] == '_': # we have to match these last
                    new_umatched.append([group, count])
                else:
                    for g,c in groups_counts.items():
                        if g[0] == group[0]:
                            if cls._atomlist_match(group[1], g[1]):
                                if count >= c:
                                    count -= c
                                    del groups_counts[g]
                                else:
                                    groups_counts[g] -= count
                                    break
                    else:
                        if count > 0:
                            return False

            for group, count in new_umatched:
                for g, c in groups_counts.items():
                    if cls._atomlist_match(group[1], g[1]):
                        if count >= c:
                            count -= c
                            del groups_counts[g]
                        else:
                            groups_counts[g] -= count
                            break
                else:
                    if count > 0:
                        return False

            return True

        return match

    ring_type_dispatch = {
            ((("C", 1),)*6): 'benzene',
            (("C", 1),)*5 +(("N", 0),): 'pyridine',
            (("C", 1), ("C", 1), ("N", 0), ("C", 1), ("C", 1), ("N", 0)): 'pyrazine',
            (("C", 1),)*4 +(("N", 0),): 'pyridine',
            # (("C", ("_",))*4, ("N",)): 'furan',
            # (("C", 5),): self._check_cycloprop_ring,
            # (("C", 5), ("N", 1)): self._check_pyridine,
            # (("C", 4), ("N", 1)): self._check_furan,
            # (("C", 4), ("N", 2)): self._check_pyrazine,
            # (("C", 3), ("N", 2)): self._check_pyrazole,
        }
    def categorize_ring(self, ring):
        ring_atoms = [self.labels[n] for n in ring]
        ring_neighbors = None
        for count_list, name in self.ring_type_dispatch.items():
            nat = len(ring_atoms)
            if len(count_list) < nat: continue
            elem_list = [c[0] for c in count_list]
            for i in range(0, nat):
                perm = [(i+j) % nat for j in range(nat)]
                rats = [ring_atoms[j] for j in perm]
                if rats == elem_list:
                    if ring_neighbors is None:
                        ring_neighbors = [
                            self._collect_neighbor_list(r, visited=set(ring))[0]
                            for r in ring
                        ]
                    # a possible choice, need to see if it matches now
                    rns = [len(ring_neighbors[j]) for j in perm]
                    if all(
                        c==t
                        for (_,c),t in zip(count_list, rns)
                    ):
                        return name, tuple(ring[p] for p in perm)
                    break
            # just to make sure
            # if (
            #         len(count_list) == len(ring_counts)
            #         and all(ring_counts.get(a, -1) == v for a,v in count_list)
            # ):
            #     match = matcher(ring)
            #     if match: return match
        return self._format_ring_counts(ring_atoms), ring

    functional_groups = {
        #TODO: replace this with RDKit identifiers
        ("C", ((("H", ()), 3),)): "methyl",
        ("C", ((("H", ()), 2),)): "ethyl",
        ("C", ((("O", ()), 1), (("O", ("H",)), 1))): "carboxylic acid",
        ("C", ((("O", ()), 2),)): "carboxylate",
        ("C", ((("O", ()), 1),)): "carboxyl",
        ("N", ((("H", ()), 3),)): "amine",
        ("N", ((("H", ()), 2),)): "amide",
        ("N", ((("O", ()), 2),)): "nitro",
        ("N", ((("O", ()), 1),)): "nitrosyl",
        ("O", (("C", 2),)): "ether",
        ("C", ((("O", ()), 1), (("O", ("C",)), 1),)): "ester",
        ("C", (("C", 2), (("O", ()), 1))): "ketone",
        ("O", ((("O", ("H",)), 1),)): "peroxide",
        ("C", ((("C", ("H", "H", "H")), 3),)): "tert-butyl"
    }
    def match_functional_group(self, root, neighbor_lists, cache=None):
        key = (root, neighbor_lists)
        if cache is not None and key in cache:
            return cache[key]
        # group_counts = itut.counts(neighbor_lists)
        # primary_counts = itut.counts(nl[0] for nl in neighbor_lists)
        matcher = self._bonding_pattern_matcher(neighbor_lists)
        for (fg_root, counts), name in self.functional_groups.items():
            if fg_root == root:
                if matcher(counts):
                # if all(
                #         (
                #             primary_counts.get(group, -1) == count
                #                 if isinstance(group, str) else
                #             group_counts.get(group, -1) == count
                #         )
                #         for group, count in counts
                # ):
                    if cache is not None:
                        cache[key] = (name, counts)
                    return (name, counts)
        else:
            if cache is not None:
                cache[key] = None
    def find_functional_groups(self):
        possible_fgs = {fg_root for (fg_root, counts), name in self.functional_groups.items()}
        fgs = []
        for n,l in enumerate(self.labels):
            if l in possible_fgs:
                neighbor_list, neighbor_inds = self._collect_neighbor_list(n, depth=2)
                match = self.match_functional_group(l, neighbor_list)
                if match:
                    name, counts = match
                    map = dict(counts)
                    inds = [n]
                    for nl,ni in zip(neighbor_list, neighbor_inds):
                        if map.get(nl, -1) > 0:
                            map[nl] -= 1
                            inds.extend(itut.flatten(ni))
                        elif map.get(nl[0], -1) > 0:
                            map[nl[0]] -= 1
                            inds.append(ni[0])
                    fgs.append([name, inds])
        return fgs


    atom_identifier = collections.namedtuple(
        "atom_identifier",
        ["ring", "group", "motif", "atom"]
    )
    def _get_identifier(self, n, label_constructor, rings, groups):
        ring = None
        group = None
        atom = self.labels[n]
        if rings is not None:
            for name,r in rings:
                if n in r:
                    ring = name
                    break
        if groups is not None:
            for name,g in groups:
                if n in g:
                    group = name
                    break
        lab = label_constructor(
            atom,
            self._collect_neighbor_list(n, depth=1)[0],
            index=n,
            graph=self
        )

        return self.atom_identifier(ring, group, lab, atom)

    def get_label_types(self,
                        label_constructor=None,
                        use_ring_identifiers=True,
                        use_functional_group_identifiers=True
                        ):
        if label_constructor is None:
            label_constructor = self._make_label

        if use_ring_identifiers:
            ring_identifiers = [
                self.categorize_ring(r)
                for r in self.rings
            ]
        else:
            ring_identifiers = None

        if use_functional_group_identifiers:
            functional_groups = self.find_functional_groups()
        else:
            functional_groups = None

        return [
            self._get_identifier(n, label_constructor, ring_identifiers, functional_groups)
            for n, _ in enumerate(self.labels)
        ]

    light_atoms = {"H", "D"}
    def get_heavy_atom_framework_graph(
            self,
            heavy_atoms=None,
            light_atoms=None,
            included_atoms=None
    ):
        if heavy_atoms is None:
            if light_atoms is None:
                light_atoms = self.light_atoms

            inds = [
                i for i,l in enumerate(self.labels)
                if l not in light_atoms
            ]
        else:
            inds = [
                i for i,l in enumerate(self.labels)
                if l in heavy_atoms
            ]
        if included_atoms is not None:
            inds = [i for i in inds if i not in inds] + inds

        return self.take(inds)
    def find_longest_chain(self,
                           rings=None,
                           root=None,
                           use_highest_valencies=True,
                           heavy_atoms=True,
                           light_atoms=None
                           ):
        if heavy_atoms or (light_atoms is not None):
            if heavy_atoms is True: heavy_atoms = None
            graph = self.get_heavy_atom_framework_graph(
                included_atoms=[root] if root is not None else None,
                heavy_atoms=heavy_atoms,
                light_atoms=light_atoms
            )
            return graph.find_longest_chain(
                root=0 if root is not None else None,
                heavy_atoms=False,
                light_atoms=None
            )
        else:
            return super().find_longest_chain()

    def segment_by_chains(self,
                          root=None,
                          rings=None,
                          use_highest_valencies=True,
                          heavy_atoms=True,
                          light_atoms=None
                          ):
        if heavy_atoms or (light_atoms is not None):
            if heavy_atoms is True: heavy_atoms = None
            graph = self.get_heavy_atom_framework_graph(
                included_atoms=[root] if root is not None else None,
                heavy_atoms=heavy_atoms,
                light_atoms=light_atoms
            )
            return graph.segment_by_chains(
                root=0 if root is not None else None,
                heavy_atoms=False,
                light_atoms=None
            )
        else:
            return super().segment_by_chains(root=root)
